import os
import sys
import re
import yaml
import logging
from collections import namedtuple
import healpy
import numpy as np
import numpy.ma as ma
import pyarrow.parquet as pq
from astropy import units as u
from desc.skycatalogs.objects.base_object import load_lsst_bandpasses
from desc.skycatalogs.objects.base_object import  CatalogContext
from desc.skycatalogs.objects import *
from desc.skycatalogs.readers import *
from desc.skycatalogs.readers import ParquetReader
from desc.skycatalogs.utils.sed_tools import ObservedSedFactory
from desc.skycatalogs.utils.sed_tools import MilkyWayExtinction
from desc.skycatalogs.utils.config_utils import Config

__all__ = ['SkyCatalog', 'open_catalog', 'Box', 'Disk', 'PolygonalRegion']


Box = namedtuple('Box', ['ra_min', 'ra_max', 'dec_min', 'dec_max'])

# radius is measured in arcseconds
Disk = namedtuple('Disk', ['ra', 'dec', 'radius_as'])

from lsst.sphgeom import ConvexPolygon, UnitVector3d, LonLat
class PolygonalRegion:

    def __init__(self, vertices_radec=None, convex_polygon=None):
        '''
        Supply either an object of type lsst.sphgeom.ConvexPolygon
        or a list of vertices, each a tuple (ra, dec) in degrees,
        which describe a convex polygon
        '''
        if convex_polygon:
            if isinstance(convex_polygon, ConvexPolygon):
                self._convex_polygon = convex_polygon
                return
        if vertices_radec:
            if not isinstance(vertices_radec, list):
                raise TypeError(f'PolygonalRegion: Argument {vertices_radec} is not a list')
            vertices = [UnitVector3d(LonLat.fromDegrees(v_rd[0], v_rd[1])) for v_rd in vertices_radec]
            self._convex_polygon = ConvexPolygon(vertices)
            return
        raise ValueError('PolygonalRegion: Either vertices_radec or convex_polygon must have an acceptable value')

    def get_vertices(self):
        '''
        Return vertices as list of 3d vectors
        '''
        return self._convex_polygon.getVertices()

    def get_vertices_radec(self):
        v3d = self.get_vertices()
        vertices_radec = []
        for v in v3d:
            vertices_radec.append((LonLat.longitudeOf(v).asDegrees(),
                                   LonLat.latitudeOf(v).asDegrees()))
        return vertices_radec

    def get_containment_mask(self, ra, dec, included=True):
        '''
        Parameters
        ----------
        ra, dec      parallel float arrays, units are degrees. Together
                     they define the list of points to be checked for
                     containment
        included     boolean   If true, mask bit will be set to True for
                               contained points, else False.    Reverse
                               the settings if included is False
        '''
        # convert to radians
        ra = [(r * u.degree).to_value(u.radian) for r in ra]
        dec = [(d * u.degree).to_value(u.radian) for d in dec]

        mask = self._convex_polygon.contains(ra, dec)
        if included:
            return mask
        else:
            return np.logical_not(mask)

# This function should maybe be moved to utils
def _get_intersecting_hps(hp_ordering, nside, region):
    '''
    Given healpixel structure defined by hp_ordering and nside, find
    all healpixels which instersect region, where region may be either
    a box or a disk.
    Return as some kind of iterable
    Note it's possible extra hps which don't actually intersect the region
    will be returned
    '''
    # First convert region description to an array (4,3) with (x,y,z) coords
    # for each vertex
    if isinstance(region, Box):
        vec = healpy.pixelfunc.ang2vec([region.ra_min, region.ra_max,
                                        region.ra_max, region.ra_min],
                                       [region.dec_min, region.dec_min,
                                        region.dec_max, region.dec_max],
                                       lonlat=True)

        return healpy.query_polygon(nside, vec, inclusive=True, nest=False)
    if isinstance(region, Disk):
        # Convert inputs to the types query_disk expects
        center = healpy.pixelfunc.ang2vec(region.ra, region.dec,
                                          lonlat=True)
        radius_rad = (region.radius_as * u.arcsec).to_value('radian')

        pixels = healpy.query_disc(nside, center, radius_rad, inclusive=True,
                                   nest=False)
        return pixels

    if isinstance(region, PolygonalRegion):
        return healpy.query_polygon(nside, region.get_vertices(),
                                    inclusive=True, nest=False)

def _compress_via_mask(tbl, id_column, region):
    '''
    Parameters
    ----------
    tbl          data table including columns named "ra", "dec", and id_column
    id_column    string
    region       mask should restrict to this region (or not at all if None)

    Returns
    -------
    4 values ra, dec, id, mask
    If objects are in the region, ra, dec, id correspond to those objects.
    mask will mask off unused objects
    If there are no objects in the region, all return values are None

    '''
    if region is not None:
        if isinstance(region, PolygonalRegion):        # special case
            # Find bounding box
            vertices = region.get_vertices_radec()  # list of (ra, dec)
            ra_max = max([v[0] for v in vertices])
            ra_min = min([v[0] for v in vertices])
            dec_max = max([v[1] for v in vertices])
            dec_min = min([v[1] for v in vertices])
            bnd_box = Box(ra_min, ra_max, dec_min, dec_max)
            # Compute mask for that box
            mask = _compute_mask(bnd_box, tbl['ra'], tbl['dec'])
            if all(mask): # even bounding box doesn't intersect table rows
                return None, None, None, None

            # Get compressed ra, dec
            ra_compress = ma.array(tbl['ra'], mask=mask).compressed()
            dec_compress = ma.array(tbl['dec'], mask=mask).compressed()
            poly_mask = _compute_mask(region, ra_compress, dec_compress)

            # Get indices of unmasked
            ixes = np.where(~mask)

            # Update bounding box mask with polygonal mask
            mask[ixes] |= poly_mask
        else:
            mask = _compute_mask(region, tbl['ra'], tbl['dec'])

        if all(mask):
            return None, None, None, None
        else:
            ra_compress = ma.array(tbl['ra'], mask=mask).compressed()
            dec_compress = ma.array(tbl['dec'], mask=mask).compressed()
            id_compress = ma.array(tbl[id_column], mask=mask).compressed()
            return ra_compress, dec_compress, id_compress, mask
    else:
        return tbl['ra'], tbl['dec'], tbl[id_column],None

def _compute_mask(region, ra, dec):
    '''
    Compute mask according to region for provided data
    Parameters
    ----------
    region         Supported shape (box, disk)  or None
    ra,dec         Coordinates for data to be masked, in degrees
    Returns
    -------
    mask of elements in ra, dec arrays to be omitted

    '''
    mask = None
    if isinstance(region, Box):
        mask = np.logical_or((ra < region.ra_min),
                             (ra > region.ra_max))
        mask = np.logical_or(mask, (dec < region.dec_min))
        mask = np.logical_or(mask, (dec > region.dec_max))
    if isinstance(region, Disk):
        # Change positions to 3d vectors to measure distance
        p_vec = healpy.pixelfunc.ang2vec(ra, dec, lonlat=True)

        c_vec = healpy.pixelfunc.ang2vec(region.ra,
                                         region.dec,
                                         lonlat=True)
        radius_rad = (region.radius_as * u.arcsec).to_value('radian')


        # Rather than comparing arcs, it is equivalent to compare chords
        # (or square of chord length)
        diff = p_vec - c_vec
        obj_chord_sq = np.sum(np.square(p_vec - c_vec),axis=1)

        # This is to be compared to square of chord for angle a corresponding
        # to disk radius.  That's 4(sin(a/2)^2)
        rad_chord_sq = 4 * np.square(np.sin(0.5 * radius_rad) )
        mask = obj_chord_sq > rad_chord_sq
    if isinstance(region, PolygonalRegion):
        mask = region.get_containment_mask(ra, dec, included=False)
    return mask

class SkyCatalog(object):
    '''
    A base class with derived classes for galaxies, static (w.r.t. coordinates)
    point sources, SSOs

    '''
    def __init__(self, config, mp=False, skycatalog_root=None, verbose=False,
                 loglevel='INFO'):
        '''
        Parameters
        ----------
        config:  dict.  Typically the result of loading a yaml file
        mp:      boolean  Default False. Set True to enable multiprocessing.
        skycatalog_root: If not None, overrides value in config or
                         in environment variable SKYCATALOG_ROOT
        '''
        self._config = Config(config)
        self._logger = logging.getLogger('skyCatalogs.client')

        if not self._logger.hasHandlers():
            self._logger.setLevel(loglevel)
            ch = logging.StreamHandler()
            ch.setLevel(loglevel)
            formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
            ch.setFormatter(formatter)
            self._logger.addHandler(ch)

        self._mp = mp
        if 'schema_version' not in config:
            self._cat_dir = config['root_directory']
        else:
            sky_root = config['skycatalog_root']        # default
            if skycatalog_root:
                sky_root = skycatalog_root
            else:
                sky_root_env = os.getenv('SKYCATALOG_ROOT', None)
                if sky_root_env:
                    sky_root = sky_root_env

            self._cat_dir = os.path.join(sky_root, config['catalog_dir'])

        self._logger.info(f'Catalog data will be read from {self._cat_dir}')
        # There may be more to do at this point but not too much.
        # In particular, don't read in anything from data files

        self.verbose = verbose
        self._validate_config()

        # Outer dict: hpid for key. Value is another dict
        #    with keys 'files', 'object_types', each with value another dict
        #    for 'files', map filepath to handle (initially None)
        #    for 'object_types', map object type to filepath
        self._hp_info = dict()
        hps = self._find_all_hps()

        self._observed_sed_factory =\
            ObservedSedFactory(self._config.get_tophat_parameters(),
                               config['Cosmology'])
        self._extinguisher = MilkyWayExtinction(self.observed_sed_factory)

        # Make our properties accessible to BaseObject, etc.
        self.catalog_context = CatalogContext(self)

    @property
    def observed_sed_factory(self):
        return self._observed_sed_factory

    @property
    def extinguisher(self):
        return self._extinguisher

    @property
    def raw_config(self):
        '''
        Return config, typically uploaded from yaml.
        '''
        return self._config

    # One might check that the config is complete and well-formed
    #  - for example require certain keys, such as catalog_name,
    #  data_file_type, area_parition, root_directory, object_types -
    # to exist, and check that that the data directory (value of
    # root_directory) exists.
    def _validate_config(self):
        pass

    def _find_all_hps(self):
        '''
        For each healpix with files matching pattern in the directory,
        update _hp_info as needed to keep track of all files for that healpix
        and the object types included in those files.

        Returns
        -------
        Set of healpix pixels with at least one file in the directory

        '''

        # If major organization is by healpix, healpix # will be in
        # subdirectory name.  Otherwise there may be no subdirectories
        # or data may be organized by component type.
        # Here only handle case where data files are directly in root dir
        files = os.listdir(self._cat_dir)
        o_types = self._config['object_types']

        hp_set = set()
        for f in files:
            for ot in o_types:
                # find all keys containing the string 'file_template'
                template_keys = [k for k in o_types[ot] if 'file_template' in k]
                for k in template_keys:
                    m = re.fullmatch(o_types[ot][k], f)
                    if m:
                        hp = int(m['healpix'])
                        hp_set.add(hp)

                        if hp not in self._hp_info:
                            self._hp_info[hp] = {'files' : {f : None},
                                                 'object_types' : {ot : [f]}}
                        else:
                            this_hp = self._hp_info[hp]
                            # Value of 'object_types' is now a list
                            if f not in this_hp['files'] :
                                this_hp['files'][f] = None
                            if ot in this_hp['object_types']:
                                this_hp['object_types'][ot].append(f)
                            else:
                                this_hp['object_types'][ot] = [f]
        return hp_set

    def get_hps_by_region(self, region):
        '''
        Parameters
        ----------
        Region can be a box (named 4-tuple (min-ra, max-ra, min-dec, max-dec)),
        a circle (named 3-tuple (ra, dec, radius)) or of type
        PolygonalRegion.
        Catalog area partition must be by healpix

        Returns
        -------
        Set of healpixels intersecting the region
        '''
        # If area_partition doesn't use healpix, raise exception

        return _get_intersecting_hps(
            self._config['area_partition']['ordering'],
            self._config['area_partition']['nside'],
            region)

    def get_object_type_names(self):
        '''
        Returns
        -------
        All object type names in the catalog's config
        '''
        return set(self._config['object_types'].keys())

    # Add more functions to return parts of config of possible interest
    # to user

    def get_objects_by_region(self, region, obj_type_set=None, datetime=None):
        '''
        Parameters
        ----------
        region         region is a named tuple(may be box or circle)
                       or object of type PolygonalRegion
        obj_type_set   Return only these objects. Defaults to all available
        datetime       Python datetime object. Ignored except for SSO

        Returns
        -------
        ObjectList containing sky objects visible in the region
        [at the specified time]
        '''
        # Take intersection of obj_type_list and available object types
        # Determine healpix intersecting the region

        if self.verbose:
            print("Region ", region)
            print("obj_type_set ", obj_type_set)
        if self._config['area_partition']['type'] == 'healpix':
            hps = self.get_hps_by_region(region)

        # otherwise raise a not-supported exception

        object_list = ObjectList()
        if obj_type_set is None:
            obj_types = self.get_object_type_names()
        else:
            obj_types = self.get_object_type_names().intersection(obj_type_set)
        for hp in hps:
            # Maybe have a multiprocessing switch? Run-time option when
            # catalog is opened?
            c = self.get_objects_by_hp(hp, region, obj_types, datetime)
            if (len(c)) > 0:
                object_list.append_object_list(c)

        return object_list

    def get_object_iterator_by_region(self, region=None, obj_type_set=None,
                                      max_chunk=None, datetime=None):
        '''
        Not yet implemented.

        Parameters
        ----------
        region         Either a box or a circle (each represented as
                       named tuple) or object of type PolygonalRegion
        obj_type_set   Return only these objects. Defaults to all available
        max_chunk      If specified, iterator will return no more than this
                       number of objections per iteration
        datetime       Python datetime object. Ignored for all but SSO

        Returns
        -------
        An iterator
        '''
        raise NotImplementedError('get_object_iterator_by_region not implemented yet. See get_objects_by_region instead for now')

    def get_objects_by_hp(self, hp, region=None, obj_type_set=None,
                          datetime=None):
        '''
        Find all objects in the healpixl (and region if specified)
        of requested types

        Parameters
        ----------
        hp     The healpix pixel.
        region If supplied defines a disk, box or convex polygon.
               Return only objects contained in it
        obj_type_set   Type of objects to fetch.  Defaults to all
        datetime       Ignored for now; no support for moving objects

        Returns
        -------
        ObjectList containing sky objects in the region and the hp
        '''
        object_list = ObjectList()

        if hp not in self._hp_info:
            print(f'WARNING: In SkyCatalog.get_objects_by_hp healpixel {hp} intersects region but has no catalog file')
            return object_list

        G_COLUMNS = ['galaxy_id', 'ra', 'dec']
        PS_COLUMNS = ['object_type', 'id', 'ra', 'dec']
        if self.verbose:
            print('Working on healpix pixel ', hp)

        obj_types = obj_type_set
        if obj_types is None:
            obj_types = self._config['object_types'].keys()
        else:
            parents = set()
            for ot in obj_types:
                if 'parent' in self._config['object_types'][ot]:
                    parents.add(self._config['object_types'][ot]['parent'])
            obj_types = obj_types.union(parents)

        # Find the right Sky Catalog file or files (depends on obj_type_set)
        # Get file handle(s) and store if we don't already have it (them)

        # Associate object types with files and files with readers.
        # May be > one type per reader (e.g. different kinds of point sources
        # stored in a single file)
        # Also may be more than one file per type (galaxy has two)
        rdr_ot = dict()   # maps readers to set of object types it reads

        for ot in obj_types:
            # FInd files associated with this type
            if 'file_template' in self._config['object_types'][ot]:
                f_list = self._hp_info[hp]['object_types'][ot]
            elif 'parent' in self._config['object_types'][ot]:
                f_list = self._hp_info[hp]['object_types'][self._config['object_types'][ot]['parent']]

            for f in f_list:
                if self._hp_info[hp]['files'][f] is None:            # no reader yet
                    full_path = os.path.join(self._cat_dir, f)
                    the_reader = parquet_reader.ParquetReader(full_path, mask=None)
                    self._hp_info[hp]['files'][f] = the_reader
                else:
                    the_reader = self._hp_info[hp]['files'][f]
                # associate object type with this reader
                if the_reader in rdr_ot:
                    rdr_ot[the_reader].add(ot)
                else:
                    rdr_ot[the_reader] = set([ot])

        # Now get minimal columns for objects using the readers
        galaxy_readers = []
        star_readers = []

        # First make a pass to separate out the readers
        for rdr in rdr_ot:
            if 'galaxy' in rdr_ot[rdr]:
                galaxy_readers.append(rdr)
            elif 'star' in rdr_ot[rdr]:
                star_readers.append(rdr)

        # Make galaxy collections
        for rdr in galaxy_readers:
            if 'ra' not in rdr.columns:
                continue

            # Make a collection for each row group

            for rg in range(rdr.n_row_groups):
                arrow_t = rdr.read_columns(G_COLUMNS, None, rg)

                ra_c, dec_c, id_c, mask = _compress_via_mask(arrow_t,
                                                             'galaxy_id',
                                                             region)

                if ra_c is not None:
                    new_collection = ObjectCollection(ra_c, dec_c, id_c,
                                                      'galaxy', hp, self,
                                                      region=region, mask=mask,
                                                      readers=galaxy_readers,
                                                      row_group=rg)
                    object_list.append_collection(new_collection)

        # Make point source collection
        for rdr in star_readers:
            if 'ra' not in rdr.columns:
                continue
            arrow_t = rdr.read_columns(PS_COLUMNS, None)

            ra_c, dec_c, id_c, mask = _compress_via_mask(arrow_t, 'id', region)

            if ra_c is not None:
                new_collection = ObjectCollection(ra_c, dec_c, id_c,
                                                  'star', hp, self,
                                                  region=region, mask=mask,
                                                  readers=star_readers)
                object_list.append_collection(new_collection)

        return object_list

    # For generator version, do this a row group at a time
    #    but if region cut leaves too small a list, read more rowgroups
    #    to achieve a reasonable size list (or exhaust the file)
    def get_object_iterator_by_hp(self, hp, obj_type_set=None,
                                  max_chunk=None, datetime=None):
        '''
        Parameters
        ----------
        datetime       Python datetime object.
        hp             A healpix id
        obj_type_set   Return only these objects. Defaults to all available
        max_chunk      If specified, iterator will return no more than this
                       number of objections per iteration
        Returns
        -------
        An iterator
        '''
        pass


def open_catalog(config_file, mp=False, skycatalog_root=None, verbose=False):
    '''
    Parameters
    ----------
    config_file   yaml file containing config
    skycatalog_root optional override of skycatalog_root. This value may also
                    be supplied by setting the environment variable
                    SKYCATALOG_ROOT.  Precedence is 1) argument
                    2) environment variable 3) value in config file for
                    key skycatalog_root.  However set, this value
                    joined to value of the key catalog_dir will be used
                    to find the catalog data.

    Returns
    -------
    SkyCatalog
    '''
    # Get LSST bandpasses in case we need to compute fluxes
    load_lsst_bandpasses()
    with open(config_file) as f:
        return SkyCatalog(yaml.safe_load(f), skycatalog_root=skycatalog_root,
                          mp=mp, verbose=verbose)

if __name__ == '__main__':
    import time
    ###cfg_file_name = 'latest.yaml'
    cfg_file_name = 'future_latest.yaml'
    skycatalog_root = os.path.join(os.getenv('SCRATCH'),'desc/skycatalogs')

    if len(sys.argv) > 1:
        cfg_file_name = sys.argv[1]
    cfg_file = os.path.join('/global/homes/j/jrbogart/desc_git/skyCatalogs/cfg',
                            cfg_file_name)

    write_sed = False
    if len(sys.argv) > 2:
        write_sed = True

    # For tract 3828
    #   55.73604 < ra < 57.563452
    #  -37.19001 < dec < -35.702481


    cat = open_catalog(cfg_file, skycatalog_root=skycatalog_root)
    hps = cat._find_all_hps()
    print('Found {} healpix pixels '.format(len(hps)))
    for h in hps: print(h)

    ra_min_tract = 55.736
    ra_max_tract = 57.564
    dec_min_tract = -37.190
    dec_max_tract = -35.702
    ##ra_min_small = 56.0
    ##ra_max_small = 56.2
    ra_min_small = 55.9
    ra_max_small = 56.1
    dec_min_small = -36.2
    dec_max_small = -36.0

    sed_fmt = 'lambda: {:.1f}  f_lambda: {:g}'

    rgn = Box(ra_min_small, ra_max_small, dec_min_small, dec_max_small)
    vertices = [(ra_min_small, dec_min_small), (ra_min_small, dec_max_small),
                (ra_max_small, dec_max_small), (ra_max_small, dec_min_small)]
    rgn_poly = PolygonalRegion(vertices_radec=vertices)

    intersect_hps = _get_intersecting_hps('ring', 32, rgn)

    print("For region ", rgn)
    print("intersecting pixels are ", intersect_hps)

    intersect_poly_hps = _get_intersecting_hps('ring', 32, rgn_poly)
    print("For region ", rgn_poly)
    print("intersecting pixels are ", intersect_poly_hps)


    print('Invoke get_objects_by_region with box region')
    t0 = time.time()
    object_list = cat.get_objects_by_region(rgn)
    t_done = time.time()
    print('Took ', t_done - t0)
                                            ##### temporary obj_type_set={'galaxy', 'star'} )
    #                                        obj_type_set=set(['galaxy']) )
    # Try out get_objects_by_hp with no region
    #colls = cat.get_objects_by_hp(9812, None, set(['galaxy']) )

    print('Number of collections returned for box:  ', object_list.collection_count)
    print('Object count for box: ', len(object_list))

    print('Invoke get_objects_by_region with polygonal region')
    t0 = time.time()
    object_list_poly = cat.get_objects_by_region(rgn_poly)
    t_done = time.time()
    print('Took ', t_done - t0)

    print('Number of collections returned for polygon:  ',
          object_list_poly.collection_count)
    assert(object_list.collection_count == object_list_poly.collection_count)
    print('Object count for polygon: ', len(object_list_poly))

    fudge = 5
    assert(len(object_list_poly) > len(object_list) - fudge)
    assert(len(object_list_poly) < len(object_list) + fudge)

    print('Now try inscribed polygon which is not a box')
    ra_avg = (ra_min_small + ra_max_small) * 0.5
    dec_avg = (dec_min_small + dec_max_small) * 0.5
    diamond_vertices = [(ra_min_small, dec_avg), (ra_avg, dec_max_small),
                        (ra_max_small, dec_avg), (ra_avg, dec_min_small)]
    rgn_diamond = PolygonalRegion(vertices_radec=diamond_vertices)

    intersect_diamond_hps = _get_intersecting_hps('ring', 32, rgn_diamond)
    print("for diamond region ", rgn_diamond)
    print("intersecting pixels are ", intersect_diamond_hps)

    print('Invoke get_objects_by_region with diamond region')
    t0 = time.time()
    object_list_diamond = cat.get_objects_by_region(rgn_diamond)
    t_done = time.time()
    print('Took ', t_done - t0)

    print('Number of collections returned for diamond:  ',
          object_list_diamond.collection_count)
    print('Object count for diamond: ', len(object_list_diamond))


    #### TEMP FOR DEBUGGING
    ### exit(0)


    colls = object_list.get_collections()
    got_a_sed = False
    for c in colls:
        n_obj = len(c)
        print("For hpid ", c.get_partition_id(), "found ", n_obj, " objects")
        print("First object: ")
        print(c[0], '\nid=', c[0].id, ' ra=', c[0].ra, ' dec=', c[0].dec,
              ' belongs_index=', c[0]._belongs_index,
              ' object_type: ', c[0].object_type )

        print("Slice [1:3]")
        slice13 = c[1:3]
        for o in slice13:
            print('id=',o.id, ' ra=',o.ra, ' dec=',o.dec, ' belongs_index=',
                  o._belongs_index,  ' object_type: ', o.object_type)
            print(o.object_type)
            if o.object_type == 'star':
                print(o.get_instcat_entry())
                #(lmbda, f_lambda, magnorm) = o.get_sed(resolution=1.0)
                sed, magnorm = o.get_sed(resolution=1.0)
                print('For star magnorm: ', magnorm)
                if magnorm < 1000:
                    print('Length of sed: ', len(sed.wave_list))
            else:
                for cmp in ['disk', 'bulge', 'knots']:
                    print(cmp)
                    if cmp in o.subcomponents:
                        print(o.get_instcat_entry(component=cmp))
                        sed, _ = o.get_sed(cmp)
                        if sed:
                            print('Length of sed table: ', len(sed.wave_list))
                            if not got_a_sed:
                                got_a_sed = True
                                th = o.get_native_attribute(f'sed_val_{cmp}')
                                print('Tophat values: ', th)
                                sed, _ = o.get_sed(component=cmp)
                                print('Simple sed wavelengths:')
                                print(sed.wave_list)
                                print('Simple sed values:')
                                print([sed(w) for w in sed.wave_list])
                                if write_sed:
                                    o.write_sed('simple_sed.txt', component=cmp)
                                sed_fine, _ = o.get_sed(component=cmp,
                                                        resolution=1.0)
                                print('Bin width = 1 nm')
                                print('Initial wl values', sed_fine.wave_list[:20])
                                print('Start at bin 100', sed_fine.wave_list[100:120])
                                print('Initial values')
                                print([sed_fine(w) for w in sed_fine.wave_list[:20]])
                                print('Start at bin 100')
                                print([sed_fine(w) for w in sed_fine.wave_list[100:120]])
                        else:
                            print('All-zero sed')

                # Try out old wrapper functions
                print("\nget_dust:")
                i_av, i_rv, g_av, g_rv = o.get_dust()
                print(f'i_av={i_av} i_rv={i_rv} g_av={g_av} g_rv={g_rv}')
                print("\nget_wl_params")
                g1, g2, mu = o.get_wl_params()
                print(f'g1={g1} g2={g2} mu={mu}')
                print("\nget_gsobject_components. Keys of returned dict:")
                gs_dict = o.get_gsobject_components()
                print(gs_dict.keys())
                print("\nget_observer_sed_components.  Keys of returned dict:")
                o_seds = o.get_observer_sed_components()
                print(o_seds.keys())

                f = o.get_LSST_flux('i')
                print(f'Flux for i bandpass: {f}')
                fluxes = o.get_LSST_fluxes()
                for k,v in fluxes.items():
                    print(f'Bandpass {k} has flux {v}')

        if n_obj > 200:
            print("Object 200")
            print(c[200], '\nid=', c[200].id, ' ra=', c[200].ra, ' dec=',
                  c[200].dec,
                  ' belongs_index=', c[200]._belongs_index)
        if n_obj > 163997:
            slice_late = c[163994:163997]
            print('\nobjects indexed 163994 through 163996')
            for o in slice_late:
                print('id=',o.id, ' ra=',o.ra, ' dec=',o.dec, ' belongs_index=',
                      o._belongs_index)

    print('Total object count: ', len(object_list))

    obj = object_list[0]
    print("Type of element in object_list:", type(obj))

    if object_list[0].object_type == 'galaxy':
        redshift0 = object_list[0].get_native_attribute('redshift')
        print('First redshift: ', redshift0)


    sum = 0
    for obj in object_list:
        sum = sum + 1
        if sum > 7216:
            print("Sum is now ", sum)
            print("obj id: ", obj.id)
        if sum > 7220:
            break

    print(f'Object list len:  {len(object_list)}')
    print(f'Objects found with "in":  {sum}')

    segment = object_list[2:5]
    print('Information for slice 2:5 ')
    for o in segment:
        print(f'object {o.id} of type {o.object_type} belongs to collection {o._belongs_to}')

    print('\nInformation for slice 285:300')
    segment = object_list[285:300]
    for o in segment:
        print(f'object {o.id} of type {o.object_type} belongs to collection {o._belongs_to}')

    #ixes = ([3,5,8],)
    ixes = (np.array([3,5,8, 300, 303]),)
    print(f'\nObjects with indexes {ixes[0]}')
    for o in object_list[ixes]:
        print(o.id)
    print(f'\nObjects in slice [3:9]')
    for o in object_list[3:9]:
        print(o.id)
    print(f'\nObjects in slice [300:304]')
    for o in object_list[300:304]:
        print(o.id)



from collections.abc import Sequence, Iterable
from collections import namedtuple
import os
import gzip
import itertools
import numpy as np
import astropy.units as u
import warnings
from dust_extinction.parameter_averages import F19
import galsim
import logging
from galsim.errors import GalSimRangeError

from desc.skycatalogs.utils.translate_utils import form_object_string
from desc.skycatalogs.utils.config_utils import Config
from desc.skycatalogs.utils.sed_tools import ObservedSedFactory

'''
Main object types.   There are also may be subtypes. For example,
there could be two subtypes for bulge components, differing in the
form of their associated SEDs
'''

__all__ = ['BaseObject', 'ObjectCollection', 'ObjectList', 'OBJECT_TYPES',
           'LSST_BANDS', 'load_lsst_bandpasses', 'CatalogContext']
GALAXY=1
GALAXY_BULGE=2
GALAXY_DISK=3
GALAXY_KNOTS=4
STAR=5
AGN=6
SN=7

OBJECT_TYPES = {'galaxy' : GALAXY, 'bulge_basic' : GALAXY_BULGE,
                'disk_basic' : GALAXY_DISK, 'knots_basic' : GALAXY_KNOTS,
                'star' : STAR, 'agn' : AGN, 'sn' : SN}

LSST_BANDS = ('ugrizy')

# global for easy access for code run within mp

class CatalogContext:
    def __init__(self, the_sky_cat):
        global sky_cat
        self._sky_cat = the_sky_cat
        sky_cat = the_sky_cat

def load_lsst_bandpasses():
    '''
    Read in lsst bandpasses from standard place, trim, and store in global dict
    Returns: The dict

    '''
    global lsst_bandpasses
    lsst_bandpasses = dict()
    rubin_sim_dir = os.getenv('RUBIN_SIM_DATA_DIR', None)
    if rubin_sim_dir:
        bp_dir = os.path.join(rubin_sim_dir, 'throughputs', 'baseline')

        if os.path.exists(bp_dir):
            BaseObject._bp_path = bp_dir
            #logger.info(f'Using rubin sim dir {rubin_sim_dir}')
        else:
            bp_dir = os.path.join(os.getenv('HOME'), 'rubin_sim_data',
                                   'throughputs', 'baseline')
            #logger.info(f'Using rubin sim dir rubin_sim_data under HOME')
            if os.path.exists(bp_dir):
                BaseObject._bp_path = bp_dir
            else:
                #logger.info('Using galsim built-in bandpasses')
                bp_dir = None
                fname_fmt = 'LSST_{band}.dat'
        for band in LSST_BANDS:
            if bp_dir:
                bp_full_path = os.path.join(bp_dir, f'total_{band}.dat')
            else:
                bp_full_path = f'LSST_{band}.dat'
            lsst_bandpasses[band] = galsim.Bandpass(bp_full_path, 'nm').thin()

    return lsst_bandpasses

class BaseObject(object):
    '''
    Abstract base class for static (in position coordinates) objects.
    Likely need a variant for SSO.
    '''

    _bp500 = galsim.Bandpass(galsim.LookupTable([499, 500, 501],[0, 1, 0]),
                             wave_type='nm').withZeropoint('AB')
    def __init__(self, ra, dec, id, object_type, redshift=None,
                 belongs_to=None, belongs_index=None):
        '''
        Minimum information needed for static (not SSO) objects
        ra, dec needed to check for region containment
        belongs_to is object collection, if any, this object belongs to
        '''
        self._ra = ra
        self._dec = dec
        self._id = id
        self._object_type = object_type
        self._redshift = redshift
        self._belongs_to = belongs_to
        self._belongs_index = belongs_index

        # All objects also include redshift information. Also MW extinction,
        # but extinction is by subcomponent for galaxies

    @property
    def ra(self):
        return self._ra

    @property
    def dec(self):
        return self._dec

    @property
    def id(self):
        return self._id

    @property
    def object_type(self):
        return self._object_type

    @property
    def redshift(self):
        if self._redshift:        return self._redshift
        if self._belongs_to:
            self._redshift = self.get_native_attribute('redshift')
        return self._redshift

    @property
    def partition_id(self):
        if self._belongs_to:
            return self._belongs_to.partition_id
        else:
            return None

    @property
    def belongs_to(self):
        return self._belongs_to

    @property
    def native_columns(self):
        '''
        Return set of all columns stored in parquet file for this object.
        May not include quantities which are constant for all objects
        of this type
        '''
        if self._belongs_to:
            return self._belongs_to.native_columns
        else:
            return None

    @property
    def subcomponents(self):
        '''
        Return list of all subcomponents (or None)
        '''
        if self._belongs_to:
            return self._belongs_to.subcomponents
        else:
            return None

    def get_native_attribute(self, attribute_name):
        if not self._belongs_to:
            raise ValueError('To fetch {attribute_name} object must be in an object collection')

        vals = self._belongs_to.get_native_attribute(attribute_name)
        return vals[self._belongs_index]

    def get_sed_metadata(self, **kwargs):
        '''
        E.g. list of wavelength or frequency intervals associated with sed values
        '''
        raise NotImplementedError

    def get_instcat_entry(self, band = 'r', component=None):
        '''
        Return the string corresponding to instance catalog line
        Parameters:
            band       One of ['u', 'g', 'r', 'i', 'z', 'y']
            component  Required iff the object has subcomponents (i.e.,
                       object type is 'galaxy')
        Returns: A string formatted like a line in an instance catalog
        '''
        if self.object_type == 'galaxy':
            if component not in self.subcomponents:
                return ''
        return form_object_string(self, band, component)

    def get_sed(self, component=None, resolution=None):
        '''
        Return sed and mag_norm for a galaxy component or for a star
        Parameters
        ----------
        component    one of 'bulge', 'disk', 'knots' for now. Other components
                     may be supported.  Ignored for stars
        resolution   desired resolution of lambda in nanometers. Ignored
                     for stars.

        Returns
        -------
        A pair (sed, mag_norm)
        '''

        if self._object_type == 'star':
            # read in the file; return data from it. Arguments are ignored
            if component:
                warnings.warn(f'get_sed: component argument ignored for object type {self._object_type}')
            if resolution:
                warnings.warn(f'get_sed: resolution argument ignored for object type {self._object_type}')
            mag_norm = self.get_native_attribute('magnorm')
            rel_path = self.get_native_attribute('sed_filepath')

            fpath = os.path.join(os.getenv('SIMS_SED_LIBRARY_DIR'),
                                 self.get_native_attribute('sed_filepath'))

            return sky_cat.observed_sed_factory.create_pointsource(rel_path), mag_norm
        if self._object_type != 'galaxy':
            raise ValueError('get_sed function only available for stars and galaxy components')
        if component not in ['disk', 'bulge', 'knots']:
            raise ValueError(f'Cannot fetch SED for component type {component}')

        th_val = self.get_native_attribute(f'sed_val_{component}')
        if  th_val is None:   #  values for this component are not in the file
            raise ValueError(f'{component} not part of this catalog')

        # if values are all zeros or nearly no point in trying to convert
        if max(th_val) < np.finfo('float').resolution:
            return None, 0.0

        z_h = self.get_native_attribute('redshift_hubble')
        z = self.get_native_attribute('redshift')

        sed = sky_cat.observed_sed_factory.create(th_val, z_h, z,
                                                     resolution=resolution)
        magnorm = sky_cat.observed_sed_factory.magnorm(th_val, z_h)

        return sed, magnorm

    def write_sed(self, sed_file_path, component=None, resolution=None):
        sed, _ = self.get_sed(component=component, resolution=None)

        wl = sed.wave_list
        flambda = [sed(w) for w in wl]

        with open(sed_file_path, 'w') as sed_file:
            for lam, flam in zip(wl, flambda):
                sed_file.write(f'{lam} {flam}\n')

    def get_dust(self):
        """Return the Av, Rv parameters for internal and Milky Way extinction."""
        internal_av = 0
        internal_rv = 1.
        galactic_av = self.get_native_attribute('MW_av')
        galactic_rv = self.get_native_attribute('MW_rv')
        return internal_av, internal_rv, galactic_av, galactic_rv

    def get_wl_params(self):
        """Return the weak lensing parameters, g1, g2, mu."""
        gamma1 = self.get_native_attribute('shear_1')
        gamma2 = self.get_native_attribute('shear_2')
        kappa =  self.get_native_attribute('convergence')
        # Compute reduced shears and magnification.
        g1 = gamma1/(1. - kappa)    # real part of reduced shear
        g2 = gamma2/(1. - kappa)    # imaginary part of reduced shear
        mu = 1./((1. - kappa)**2 - (gamma1**2 + gamma2**2)) # magnification
        return g1, g2, mu

    def get_gsobject_components(self, gsparams=None, rng=None):
        """
        Return a dictionary of the GSObject components for the
        sky catalogs object, keyed by component name.
        """
        if gsparams is not None:
            gsparams = galsim.GSParams(**gsparams)
        if self.object_type == 'star':
            return {None: galsim.DeltaFunction(gsparams=gsparams)}
        if self.object_type != 'galaxy':
            raise RuntimeError("Do not know how to handle object type "
                               f"{self.object_type}")
        obj_dict = {}
        for component in self.subcomponents:
            # knots use the same major/minor axes as the disk component.
            my_component = 'disk' if component != 'bulge' else 'bulge'
            a = self.get_native_attribute(
                f'size_{my_component}_true')
            b = self.get_native_attribute(
                f'size_minor_{my_component}_true')
            assert a >= b
            pa = self.get_native_attribute('position_angle_unlensed')
            beta = float(90 + pa)*galsim.degrees
            hlr = (a*b)**0.5   # approximation for half-light radius
            if component == 'knots':
                npoints = self.get_native_attribute('n_knots')
                assert npoints > 0
                obj = galsim.RandomKnots(npoints=npoints,
                                         half_light_radius=hlr, rng=rng,
                                         gsparams=gsparams)
            else:
                n = self.get_native_attribute(f'sersic_{component}')
                # Quantize the n values at 0.05 so that galsim can
                # possibly amortize sersic calculations from the previous
                # galaxy.
                n = round(n*20.)/20.
                obj = galsim.Sersic(n=n, half_light_radius=hlr,
                                    gsparams=gsparams)
            shear = galsim.Shear(q=b/a, beta=beta)
            obj = obj._shear(shear)
            g1, g2, mu = self.get_wl_params()
            obj_dict[component] = obj._lens(g1, g2, mu)
        return obj_dict

    def get_observer_sed_component(self, component):
        """
        Return the SED for the specified subcomponent of the SkyCatalog
        object, applying internal extinction, redshift, and Milky Way
        extinction.

        For Milky Way extinction, the Fitzpatrick, et al. (2019) (F19)
        model, as implemented in the dust_extinction package, is used.

        The SEDs are computed assuming exposure times of 1 second.
        """
        # Create a galsim.SED for this subcomponent.
        if self._object_type == 'star':   # or other pointsource
            sed, magnorm = self.get_sed(component=component)
            # The SED is in units of photons/nm/cm^2/s
            # -0.9210340371976184 = -np.log(10)/2.5. Use to convert mag to flux
            flux_500 = np.exp(-0.9210340371976184 * magnorm)
            sed = sed.withMagnitude(0, self._bp500)
            sed = sed*flux_500
        else:
            # For galaxy components sed already has correct normalization
            sed, _ = self.get_sed(component=component)
            if sed is None:
                # This subcomponent has zero emission so return None.
                return None


        iAv, iRv, mwAv, mwRv = self.get_dust()
        if iAv > 0:
            # Apply internal extinction model, which is assumed
            # to be the same for all subcomponents.
            pass  #TODO add implementation for internal extinction.

        # redshift already applied by create or create_pointsource

        # Apply Milky Way extinction.
        sed = sky_cat.extinguisher.extinguish(sed, mwAv)
        return sed

    def get_observer_sed_components(self):
        """
        Return a dictionary of the SEDs, keyed by component name.
        """
        sed_components = {}
        subcomponents = [None] if not self.subcomponents \
            else self.subcomponents
        for component in subcomponents:
            sed = self.get_observer_sed_component(component)
            if sed is not None:
                sed_components[component] = sed
        return sed_components

    def get_total_observer_sed(self):
        """
        Return the SED summed over SEDs for the individual SkyCatalog
        components.
        """
        sed = None
        for sed_component in self.get_observer_sed_components().values():
            if sed is None:
                sed = sed_component
            else:
                sed += sed_component
        if 'shear_1' in self.native_columns:
            _, _, mu = self.get_wl_params()
            sed *= mu
        return sed

    def get_flux(self, bandpass, sed=None):
        """
        Return the total object flux integrated over the bandpass
        in photons/sec/cm^2
        Use supplied sed if there is one
        """

        if not sed:
            sed = self.get_total_observer_sed()

        flux = sed.calculateFlux(bandpass)

        return flux

    def get_fluxes(self, bandpasses):
        # To avoid recomputing sed
        sed = self.get_total_observer_sed()
        return [sed.calculateFlux(b) for b in bandpasses]

    def get_LSST_flux(self, band, sed=None, cache=True):
        if not band in LSST_BANDS:
            return None
        att = f'lsst_flux_{band}'
        # Check if it's already an attribute
        val = getattr(self, att, None)
        if val is not None:
            return val

        if att in self.native_columns:
            return self.get_native_attribute(att)

        val = self.get_flux(lsst_bandpasses[band], sed=sed)

        if cache:
            setattr(self, att, val)
        return val

    def get_LSST_fluxes(self, cache=True, as_dict=True):
        '''
        Return a dict of fluxes for LSST bandpasses or, if as_dict is False,
        just a list in the same order as LSST_BANDS
        '''
        fluxes = dict()
        sed = self.get_total_observer_sed()
        for band in LSST_BANDS:
            ##### temporary to debug error, now apparently fixed
            try:
                fluxes[band] = self.get_LSST_flux(band, sed=sed, cache=cache)
            except GalSimRangeError as e:
                logger = logging.getLogger('skyCatalogs.creator')
                galaxy_id = self.get_native_attribute('galaxy_id')
                logger.debug(f'galaxy_id: {galaxy_id}  band: {band}')
                redshift = self.get_native_attribute('redshift')
                logger.debug(f'redshift: {redshift}')
                logger.debug(f'GalSimRangeError: {e}')
                fluxes[band] = 0.0

        if as_dict:
            return fluxes
        else:
            return list(fluxes.values())

class ObjectCollection(Sequence):
    '''
    Class for collection of static objects coming from the same
    source (e.g., file for particular healpix)
    Many of the methods look the same as for BaseObject but they return arrays
    rather than a single number.  There are some additional methods
    '''
    def __init__(self, ra, dec, id, object_type, partition_id, sky_catalog,
                 region=None, mask=None, readers=None, row_group=0):
        '''
        Minimum information needed for static objects.
        specified, has already been used by the caller to generate mask.
        ra, dec, id must be array-like and the same length
        object_type  may be either single value or array-like (could be
        useful if more than one object type is stored in the same file)
        partition_id (e.g. healpix id)
        sky_catalog instance of SkyCatalog class
        (redshift should probably be ditched; no need for it)
        (similarly for region with current code structure. Information
        needed is encoded in mask)

        mask  indices to be masked off, e.g. because region does not
              include the entire healpix pixel
        readers one per file associated with the type(s). (Galaxy info
              is stored in two files)


        '''
        self._ra = np.array(ra)
        self._dec = np.array(dec)
        self._id = np.array(id)
        self._rdrs = readers
        self._partition_id = partition_id
        self._sky_catalog = sky_catalog   # might not need this
        self._config = Config(sky_catalog.raw_config)
        self._mask = mask
        self._row_group = row_group

        # Maybe the following is silly and object_type should always be stored
        # as arrays
        if isinstance(object_type, list):
            self._object_types = np.array(object_type)
            self._object_type_unique = None
            self._uniform_object_type = False
        else:
            self._object_types = None
            self._object_type_unique = object_type
            self._uniform_object_type = True

        self._region = region

    @property
    def partition_id(self):
        return self._partition_id

    @property
    def config(self):
        return self._config

    @property
    def sky_catalog(self):
        return self._sky_catalog

    @property
    def native_columns(self):
        '''
        Return set of all columns stored for objects in this collection
        May not include quantities which are constant for all objects
        of this type
        '''
        columns = set()
        for rdr in self._rdrs:
            columns = columns.union(rdr.columns)
        return columns

    @property
    def subcomponents(self):
        '''
        Return list of all subcomponents for objects in this collection.
        Only galaxies have subcomponents
        '''
        subs = []
        if self._object_type_unique == 'galaxy':
            for s in ['bulge', 'disk', 'knots']:
                if f'sed_val_{s}' in self.native_columns:
                    subs.append(s)
        return subs

    def get_native_attribute(self, attribute_name):
        '''
        Retrieve a particular attribute for a collection
        If we already have it, just return it.  Otherwise attempt
        to fetch.   Reader should check whether the attribute actually
        exists.
        '''
        val = getattr(self, attribute_name, None)
        if val is not None: return val

        for r in self._rdrs:
            if attribute_name in r.columns:
                d = r.read_columns([attribute_name], self._mask,
                                   row_group=self._row_group)
                if not d:
                    return None
                val = d[attribute_name]
                setattr(self, attribute_name, val)
                return val

    def get_native_attributes(self, attribute_list):
        '''
        Return requested attributes as dict. Keys are column names.
        Use our mask if we have one
        '''
        remaining = set(attribute_list)
        final_d = dict()
        for r in self._rdrs:
            to_read = remaining.intersection(r.columns)
            d = r.read_columns(to_read, self._mask)
            remaining.difference_update(to_read)
            for k in d:
                final_d[k] = d[k]

        if len(remaining) > 0:
            # raise exception?   log error?
            print(f'Unknown column or columns {remaining}')
            return None

        return final_d

    def get_native_attributes_iterator(self, attribute_names):
        '''
        Return iterator for  list of attributes for a collection.  Most of
        the work probably happens in the Parquet reader

        Parameters
        ----------
        attribute_names  list of attribute names
        row_group

        Returns
        -------
        iterator which returns df for a chunk of values of the attributes
        '''
        pass        #    for now

    # implement Sequence methods
    def __contains__(self, obj):
        '''
        Parameters
        ----------
        obj can be an (object id) or of type BaseObject
        '''
        if type(obj) == type(10):
            id = obj
        else:
            if isinstance(obj, BaseObject):
                id = obj.id
            else:
                raise TypeError
        return id in self._id

    def __len__(self):
        return len(self._id)

    #def __iter__(self):   Standard impl based on __getitem__ should be ok
    #def __reversed__(self):   Default implementation ok

    def __getitem__(self, key):
        '''
        Parameters
        ----------
        If key is an int (or int-like) return a single base object
        If key is a slice return a list of objects
        If key is a tuple where key[0] is iterable of int-like,
           return a list of objects
        '''

        if self._uniform_object_type:
            object_type = self._object_type_unique
        else:
            object_type = self._object_types[key]

        if isinstance(key, int) or isinstance(key, np.int64):
            return BaseObject(self._ra[key], self._dec[key], self._id[key],
                              object_type, belongs_to=self,
                              belongs_index=key)

        elif type(key) == slice:
            ixdata = [i for i in range(min(key.stop,len(self._ra)))]
            ixes = itertools.islice(ixdata, key.start, key.stop, key.step)
            return [BaseObject(self._ra[i], self._dec[i], self._id[i],
                               object_type, belongs_to=self, belongs_index=i)
                    for i in ixes]

        elif type(key) == tuple and isinstance(key[0], Iterable):
            #  check it's a list of int-like?
            return [BaseObject(self._ra[i], self._dec[i], self._id[i],
                               object_type, belongs_to=self, belongs_index=i)
                    for i in key[0]]


    def get_partition_id(self):
        return self._partition_id

    def count(self, obj):
        '''
        returns # of occurrences of obj.  It can only be 0 or 1
        '''
        if self.__contains__(obj): return 1
        return 0

    def index(self, obj):
        '''
        Use object id to find the index
        '''
        return self._id.index(obj.id)

LocatedCollection = namedtuple('LocatedCollection',
                               ['collection', 'first_index', 'upper_bound'])
'''
Describes collection included in a list of collections to be concatenated
   collection:     the data
   first_index: index in the concatenated list of first elt in this collection
   upper_bound: index + 1 in concatenated list of last elt in this collection
      so upper_bound for one collection = first_index in the next
'''

class ObjectList(Sequence):
    '''
    Keep track of a list of ObjectCollection objects, but from user
    perspective appears to be a list of objects.  Make use of ObjectCollection
    implementation of Sequence until it's time to move to the next
    ObjectCollection
    '''

    def __init__(self):
        # Elements of _located are named tuples:
        #   (collection, first_index, upper_bound)
        self._located = []
        self._total_len = 0

    def append_collection(self, coll):
        old = self._total_len
        self._total_len += len(coll)
        self._located.append(LocatedCollection(coll, old, self._total_len))

    def append_object_list(self, object_list):
        for e in object_list._located:
            self.append_collection(e.collection)

    def redshifts(self):
        return self.get_native_attribute('redshift')

    def get_native_attribute(self, attribute_name):
        '''
        Retrieve a particular attribute for a source.
        The work is delegated to each of the constituent collections
        '''
        val = self._located[0].collection.get_native_attribute(attribute_name)
        for c in self._located[1:]:
            val = np.append(val,
                            c.collection.get_native_attribute(attribute_name))

        return val

    @property
    def collection_count(self):
        return len(self._located)

    def get_collections(self):
        '''
        return constituent ObjectCollection objects in a list
        '''
        collections = []
        for  e in self._located:
            collections.append(e.collection)

        return collections


    # implement Sequence methods
    def __contains__(self, obj):
        '''
        Parameters
        ----------
        obj can be an (object id) or of type BaseObject
        '''

        for e in self._located:
            if obj in e.collection:
                return True

        return False

    def __len__(self):
        return self._total_len

    #def __iter__(self):   Standard impl based on __getitem__ should be ok??
    #def __reversed__(self):   Default implementation ok??

    def __getitem__(self, key):
        '''
        Parameters
        ----------
        If key is an int return a single base object
        If key is a slice return a list of object
        '''
        one_only = isinstance(key, int) or isinstance(key, np.int64)
        is_slice = type(key) == slice
        is_list = type(key) == tuple and isinstance(key[0], Iterable)

        if one_only:
            start = key
        elif is_slice:
            start = key.start
        elif is_list:
            start_ix = 0
            key_list = key[0]
            start = key_list[start_ix]
        else:
            raise ValueError(f'Unknown key type {type(key)}')

        to_return = []
        for e in self._located:
            if start >= e.first_index and start < e.upper_bound:
                rel_first_index = start - e.first_index
                if one_only:
                    return e.collection[rel_first_index]
                if is_slice:
                    if key.stop < e.upper_bound:
                        rel_stop_ix = key.stop - e.first_index
                        to_return += e.collection[slice(rel_first_index,
                                                        rel_stop_ix)]
                        return to_return
                    else:
                        to_return += e.collection[slice(rel_first_index, len(e.collection))]
                    start = e.upper_bound
                if is_list:
                    sub = [elem - e.first_index for elem in key_list if elem >= e.first_index and elem < e.upper_bound]
                    to_return += e.collection[(sub,)]

                    start_ix +=len(sub)
                    if start_ix >= len(key_list):
                        break
                    start = key_list[start_ix]

        if len(to_return):
            return to_return
        else:
            raise IndexError



from collections import OrderedDict
import pyarrow as pa
import pyarrow.parquet as pq
import numpy as np
import numpy.ma as ma
import warnings

__all__ = ['ParquetReader']

class ParquetReader:
    '''
    Handle reads for a particular file
    Create a separate ParquetReader instance for each file, region
    (as specified by optional mask argument)
    '''
    def __init__(self, filepath, mask=None):
        self._filepath = filepath
        # meta data includes num_columns, num_rows, num_row_groups
        self._meta = pq.read_metadata(filepath)
        self._schema = pq.read_schema(filepath)
        self._columns = set(self._schema.names)
        #  To support iterator also save # row groups, length of each.
        #  If mask, must be able to get slices corresponding to the different
        #  row groups
        self._open()

    def _open(self):
        self._pqfile = pq.ParquetFile(self._filepath)

    def close(self):
        '''
        There doesn't seem to be any explicit way to close the file
        via pyarrow.parquet.  Maybe delete the object?   Until this is
        understood do nothing.
        '''

        if self._pqfile:
            pass

    @property
    def columns(self):
        return self._columns

    @property
    def n_row_groups(self):
        return self._meta.num_row_groups

    def read_columns(self, cols, mask, row_group=-1):
        '''
        Parameters
        -----------
        cols       list of column names belonging to the file
        mask       if not None, use it and return compressed array

        Returns
        -------
        OrdereDict where keys are column names, values are numpy arrays
        '''

        if not set(cols).issubset(self._columns):
            unknown = set(cols) - self._columns
            # raise exception?   For now, just
            print(f'Unknown column or columns {unknown}')
            return None

        d = OrderedDict()
        if not self._pqfile:
            self._open()
        if row_group < 0:
            tbl = self._pqfile.read(columns=cols)
        else:
            tbl = self._pqfile.read_row_group(row_group, columns=cols)
        for c in cols:
            c_data = np.array(tbl[c])
            if mask is not None:
                c_data = ma.array(c_data, mask=mask).compressed()

            d[c] = np.array([_ for _ in c_data])

        return d



from .base_object import BaseObject, GALAXY, GALAXY_BULGE, GALAXY_DISK, GALAXY_KNOTS
def class GalaxyObject(BaseObject):
'''
Galaxy object.
'''
      def __init__(self, ra, dec, id, redshift=None, hp_id=None,
                   belongs_to=None, belongs_index=None)
        '''
        Minimum information needed for position static (not SSO) objects
        Type of object_type should perhaps be an enumeration
        '''
        super().__init__(ra, dec, id, BaseObject.GALAXY, redshift,
                         hp_id, belongs_to, belongs_index)

        self._cmps = {GALAXY_BULGE : GalaxySub(self, GALAXY_BULGE),
                      GALAXY_DISK : GalaxySub(self, GALAXY_DISK)}
        # Don't always have knots.  When we do, handling could be different
        #               GALAXY_KNOTS : GalaxySub(self, GALAXY_KNOTS)
        #    or         GALAXY_KNOTS : KnotsObject(self)

    def get_subcomponent(self, cmp_type):
        return self._cmps.get(cmp_type)

    def get_bulge_magnorm(self):
        if self._bulge_magnorm:
            return self._bulge_magnorm
        if self._belongs_to:
            self._bulge_magnorm = self._belongs_to.redshifts()[self._belongs_index]
            return self._bulge_magnorm

    def get_disk_magnorm(self):
        if self._disk_magnorm:
            return self._disk_magnorm
        if self._belongs_to:
            self._disk_magnorm = self._belongs_to.redshifts()[self._belongs_index]
        return self._disk_magnorm

    ##### -------------------------------------------
    # commented stuff below needs thought, likely redesign. Or, in the
    # case of get_flux, we probably just get rid of it

    # def get_flux(self, date_time, band):
    #     '''
    #     Parameters
    #     ----------
    #     date_time   datetime object; time at which flux is requested
    #     band        specifies interval over which flux is to be integrated
    #                 (and filter characteristics?)
    #     '''
    #     raise NotImplementedError

    # def get_sed(self, subcomponent_list=[GALAXY_BULGE, GALAXY_DISK, GALAXY_KNOTS], **kwargs):
    #     '''
    #     For galaxies may want to specify subcomponent(s)
    #     '''
    #     d = {}
    #     for c in subcomponent_list:
    #         d[c] = self._cmps[c].get_sed(**kwargs)

    #     return d

    # def get_sed_metadata(self, **kwargs):
    #     '''
    #     Returns, e.g,. units (wavelength or frequency) and list intervals associated
    #     with sed values
    #     '''
    #     d = {}
    #     for c in subcomponent_list:
    #         d[c] = self._cmps[c].get_sed_metadata(**kwargs)

    #     return d

# All galaxy information which is not indirected (e.g. if SEDs are stored
# in files there will be separate ones for different components) is read
# in at once. Provide a way to restrict information returned to a particular
# component if desired


def class GalaxySub(object):
    def __init__(self, parent, component_type):
        self._parent = parent
        self._cmp = component_type

    @property
    def ra(self):
        return self._parent.ra

    @property
    def dec(self):
        return self._parent.dec

    @property
    def id(self):
        return self._parent.id

    @property
    def object_type(self):
        return self._cmp

    @property
    def redshift(self):
        return self._parent.redshift

    def get_magnorm(self):
        if self._cmp == GALAXY_BULGE:
            return self._parent.get_bulge_magnorm()
        if self._cmp == GALAXY_DISK:
            return self._parent.get_disk_magnorm()

    def get_sed(self):
        pass

    def

def class GalaxyKnots(object):
    def __init__(self, parent):
        self._parent = parent
        self._cmp = GALAXY_KNOTS



import io
import argparse

'''
Read in a bit of old instance catalogs (single healpixel; bulge, disk and
knots).  For each of the three:
1. Transpose; store in dataframe. Columns are
prefix (string 'object')                           ignore
id     (identifies galaxy)                         use
ra,dec                                             use
magnitude   (for band observed in this visit)        ??
sedfilepath                                         probably ignore
redshift                                            use
gamma1,gamma2,kappa                                 use
raOffset,decOffset                                  ignore
spatialmodel                                        goes in config
majorAxis,minorAxis,positionAngle,sindex            use (spatial_params)
internalExtinctionModel                             goes in config
internalAv,internalRv                               use
galacticExtinctionModel                             goes in config
galacticAv,galacticRv                               use

2. Add column for component type
3. Add a column for tophat values (make them up?)
4. Add empty column for rel. SED filepath(s)
5. Sort by galaxy id
6. write out as parquet

'''



import os
import re
from astropy import units as u
from astropy.cosmology import FlatLambdaCDM
import astropy.constants
import pandas as pd

import numpy as np
import numpy.ma as ma
from numpy.random import default_rng
from dust_extinction.parameter_averages import F19
import galsim

__all__ = ['ObservedSedFactory', 'MilkyWayExtinction', 'AB_mag',
           'get_star_sed_path']
class ObservedSedFactory:
    _clight = astropy.constants.c.to('m/s').value
    # Conversion factor below of cosmoDC2 tophat Lnu values to W/Hz comes from
    # https://github.com/LSSTDESC/gcr-catalogs/blob/master/GCRCatalogs/SCHEMA.md
    _to_W_per_Hz = 4.4659e13

    def __init__(self, th_definition, cosmology, delta_wl=0.001):
        # Get wavelength and frequency bin boundaries.
        bins = th_definition
        wl0 = [_[0] for _ in bins]

        # Find index of original bin which includes 500 nm == 5000 ang
        ix = -1
        for w in wl0:
            if w > 5000:
                break
            ix += 1

        self._ix_500nm = ix

        wl0.append(bins[-1][0] + bins[-1][1])

        wl0 = 0.1*np.array(wl0)
        self.wl = np.array(wl0)
        self.nu = self._clight/(self.wl*1e-9)  # frequency in Hz

        # Also save version of wl where vertical rise is replaced by
        # steep slope
        wl_deltas = []
        for i in range(len(bins)):
            wl_deltas.extend((self.wl[i], self.wl[i+1] - delta_wl))

        # Prepend more bins which will have 0 value
        n_bins = int(wl0[0]) - 1
        pre_wl = [float(i) for i in range(n_bins)]

        wl_deltas = np.insert(wl_deltas, 0, pre_wl)

        # Also make a matching array of 0 values
        self.pre_val = [0.0 for i in range(n_bins)]

        self._wl_deltas = wl_deltas
        self._wl_deltas_u_nm = wl_deltas*u.nm

        # Create a FlatLambdaCDM cosmology from a dictionary of input
        # parameters.  This code is based on/borrowed from
        # https://github.com/LSSTDESC/gcr-catalogs/blob/master/GCRCatalogs/cosmodc2.py#L128
        cosmo_astropy_allowed = FlatLambdaCDM.__init__.__code__.co_varnames[1:]
        ##cosmo_astropy = {k: v for k, v in config['Cosmology'].items()
        ##                 if k in cosmo_astropy_allowed}
        cosmo_astropy = {k: v for k, v in cosmology.items()
                         if k in cosmo_astropy_allowed}
        self.cosmology = FlatLambdaCDM(**cosmo_astropy)

        self.sims_sed_library_dir = os.getenv('SIMS_SED_LIBRARY_DIR')

    # Useful for getting magnorm from f_nu values
    @property
    def ix_500nm(self):
        return self._ix_500nm

    @property
    def wl_deltas(self):
        return self._wl_deltas

    @property
    def wl_deltas_u_nm(self):
        return self._wl_deltas_u_nm

    def dl(self, z):
        """
        Return the luminosity distance in units of meters.
        """
        # Conversion factor from Mpc to meters (obtained from pyccl).
        MPC_TO_METER = 3.085677581491367e+22
        return self.cosmology.luminosity_distance(z).value*MPC_TO_METER

    def create(self, Lnu, redshift_hubble, redshift, resolution=None):
        '''
        Given tophat values from cosmoDC2 produce redshifted sed.
        Does not apply extinction.
        '''
        # Compute Llambda in units of W/nm
        Llambda = (Lnu*self._to_W_per_Hz*(self.nu[:-1] - self.nu[1:])
                   /(self.wl[1:] - self.wl[:-1]))

        # Fill the arrays for the galsim.LookupTable.   Prepend
        # zero-valued bins down to mix extinction wl to handle redshifts z > 2.
        my_Llambda = []
        my_Llambda += self.pre_val
        for i in range(len(Llambda)):
            # Dealt with wl already in __init__
            my_Llambda.extend((Llambda[i], Llambda[i]))

        # Convert to (unredshifted) flux given redshift_hubble.
        flambda = np.array(my_Llambda)/(4.0*np.pi*self.dl(redshift_hubble)**2)

        # Convert to cgs units
        flambda *= (1e7/1e4)  # (erg/joule)*(m**2/cm**2)

        # Create the lookup table.
        lut = galsim.LookupTable(self.wl_deltas, flambda, interpolant='nearest')

        if resolution:
            wl_min = min(self.wl_deltas)
            wl_max = max(self.wl_deltas)
            wl_res = np.linspace(wl_min, wl_max, int((wl_max - wl_min)/resolution))
            flambda_res = [lut(wl) for wl in wl_res]
            lut = galsim.LookupTable(wl_res, flambda_res, interpolant='linear')

        # Create the SED object and apply redshift.
        sed = galsim.SED(lut, wave_type='nm', flux_type='flambda')\
                    .atRedshift(redshift)

        return sed

    def create_pointsource(self, rel_path, redshift=0):
        '''
        Return a galsim SED from information in a file
        '''
        fpath = os.path.join(self.sims_sed_library_dir, rel_path)

        sed = galsim.SED(fpath, wave_type='nm', flux_type='flambda')
        if redshift > 0:
            sed = sed.atRedshift(redshift)
        return sed

    def magnorm(self, tophat_values, z_H):
        one_Jy = 1e-26  # W/Hz/m**2
        Lnu = tophat_values[self.ix_500nm]*self._to_W_per_Hz  # convert to W/Hz
        Fnu = Lnu/4/np.pi/self.dl(z_H)**2
        with np.errstate(divide='ignore', invalid='ignore'):
            return -2.5*np.log10(Fnu/one_Jy) + 8.90

class MilkyWayExtinction:
    '''
    Applies extinction to a SED
    '''
    def __init__(self, sed_factory, ext_bin_width=0.1, mwRv=3.1):
        '''
        sed_factory         instance of ObservedSedFactory

        '''

        # Save bins to be used for extinction. F19x_range is in units of
        # 1/micron so convert.
        wl_ext_min = 1e3/F19.x_range[1]
        wl_ext_max = 1e3/F19.x_range[0]
        wl_extinct = np.linspace(wl_ext_min, wl_ext_max,
                                 int((wl_ext_max - wl_ext_min)/ext_bin_width))
        self.wl_extinct = wl_extinct
        self.wl_extinct_nm = wl_extinct*u.nm
        self.extinction = F19(Rv=mwRv)

        my_wl = sed_factory.wl_deltas
        my_wl = my_wl[np.where((wl_ext_min < my_wl) & (my_wl < wl_ext_max))]
        self.wl_deltas = my_wl
        self.wl_deltas_u = my_wl*u.nm

    def extinguish(self, sed, mwAv):
        ext = self.extinction.extinguish(self.wl_deltas_u, Av=mwAv)
        spec = galsim.LookupTable(self.wl_deltas, ext, interpolant='linear')
        mw_ext = galsim.SED(spec, wave_type='nm', flux_type='1')
        sed = sed*mw_ext

        return sed

class AB_mag:
    """
    Convert flux to AB magnitude for a set of bandpasses.
    """
    def __init__(self, bps):
        ab_sed = galsim.SED(lambda nu : 10**(8.90/2.5 - 23), wave_type='nm',
                            flux_type='fnu')
        self.ab_fluxes = {band: ab_sed.calculateFlux(bp) for
                          band, bp in bps.items()}
    def __call__(self, flux, band):
        return -2.5*np.log10(flux/self.ab_fluxes[band])

_standard_dict = {'lte' : 'starSED/phoSimMLT',
                  'bergeron' : 'starSED/wDs',
                  'km|kp' : 'starSED/kurucz'}

def get_star_sed_path(filename, name_to_folder=_standard_dict):
    '''
    Return numpy array of full paths relative to SIMS_SED_LIBRARY_DIR,
    given filenames

    Parameters
    ----------
    filename       list of strings. Usually full filename but may be missing final ".gz"
    name_to_folder dict mapping regular expression (to be matched with
                   filename) to relative path for containing directory

    Returns
    -------
    Full path for file, relative to SIMS_SED_LIBRARY_DIR
    '''

    compiled = { re.compile(k) : v for (k, v) in name_to_folder.items()}

    path_list = []
    for f in filename:
        m = None
        matched = False
        for k,v in compiled.items():
            f = f.strip()
            m = k.match(f)
            if m:
                p = os.path.join(v, f)
                if not p.endswith('.gz'):
                    p = p + '.gz'
                path_list.append(p)
                matched = True
                break

        if not matched:
            raise ValueError(f'get_star_sed_path: Filename {f} does not match any known patterns')
    return np.array(path_list)



import os
import sys
import re
import math
import logging
import yaml
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from multiprocessing import Process, Pipe
from astropy.coordinates import SkyCoord
import sqlite3
from desc.skycatalogs.utils.common_utils import print_date
from desc.skycatalogs.utils.sed_tools import ObservedSedFactory, get_star_sed_path
from desc.skycatalogs.utils.config_utils import create_config, assemble_SED_models
from desc.skycatalogs.utils.config_utils import assemble_MW_extinction, assemble_cosmology, assemble_object_types, assemble_provenance, write_yaml
from desc.skycatalogs.utils.parquet_schema_utils import make_galaxy_schema, make_galaxy_flux_schema, make_star_schema, make_star_flux_schema
from desc.skycatalogs.objects.base_object import LSST_BANDS, BaseObject

from desc.skycatalogs.objects.base_object import ObjectCollection

# from dm stack
from dustmaps.sfd import SFDQuery

"""
Code to create a sky catalog for particular object types
"""

#####pixels = [9556, 9683, 9684, 9812, 9813, 9940]

__all__ = ['CatalogCreator']

_MW_rv_constant = 3.1


_Av_adjustment = 2.742
def _make_MW_extinction(ra, dec):
    '''
    Given arrays of ra & dec, create a MW Av column corresponding to V-band
    correction.
    See "Plotting Dust Maps" example in
    https://dustmaps.readthedocs.io/en/latest/examples.html

    The coefficient _Av_adjustment comes Table 6 in Schlafly & Finkbeiner (2011)
    See http://iopscience.iop.org/0004-637X/737/2/103/article#apj398709t6

    Parameters
    ----------
    ra, dec - arrays specifying positions where Av is to be computed
    Return:
    Array of Av values
    '''

    sfd = SFDQuery()
    ebv_raw = np.array(sfd.query_equ(ra, dec))

    return _Av_adjustment * ebv_raw

def _generate_sed_path(ids, subdir, cmp):
    '''
    Generate paths (e.g. relative to SIMS_SED_LIBRARY_DIR) for galaxy component
    SED files
    Parameters
    ----------
    ids        list of galaxy ids
    subdir    user-supplied part of path
    cmp      component for which paths should be generated

    returns
    -------
    A list of strings.  The entries in the list have the form
    <subdir>/<cmp>_<id>.txt
    '''
    r = [f'{subdir}/{cmp}_{id}.txt' for id in ids]
    return r

# Collection of galaxy objects for current row group, current pixel
# Used while doing flux computation

def _do_galaxy_flux_chunk(send_conn, galaxy_collection, l_bnd, u_bnd):
    '''
    output connection
    l_bnd, u_bnd     demarcates slice to process

    returns
                    dict with keys id, lsst_flux_u, ... lsst_flux_y
    '''
    out_dict = {}

    o_list = galaxy_collection[l_bnd : u_bnd]
    out_dict['galaxy_id'] = [o.get_native_attribute('galaxy_id') for o in o_list]
    all_fluxes =  [o.get_LSST_fluxes(as_dict=False) for o in o_list]
    all_fluxes_transpose = zip(*all_fluxes)
    for i, band in enumerate(LSST_BANDS):
        v = all_fluxes_transpose.__next__()
        out_dict[f'lsst_flux_{band}'] = v

    if send_conn:
        send_conn.send(out_dict)
    else:
        return out_dict

class CatalogCreator:
    def __init__(self, parts, area_partition, skycatalog_root=None,
                 catalog_dir='.', galaxy_truth=None,
                 config_path=None, catalog_name='skyCatalog',
                 output_type='parquet', mag_cut=None,
                 sed_subdir='galaxyTopHatSED', knots_mag_cut=27.0,
                 knots=True, logname='skyCatalogs.creator',
                 pkg_root=None, skip_done=False, flux_only=False,
                 main_only=False, flux_parallel=16, provenance=None):
        """
        Store context for catalog creation

        Parameters
        ----------
        parts           Segments for which catalog is to be generated. If
                        partition type is HEALpix, parts will be a collection
                        of HEALpix pixels
        area_partition  Dict characterizing partition; e.g. HEALpix,
                        nside=<something>
        skycatalog_root Typically absolute directory containing one or
                        more subdirectories for sky catalogs. Defaults
                        to current directory
        catalog_dir     Directory relative to skycatalog_root where catalog
                        will be written.  Defaults to '.'
        galaxy_truth    GCRCatalogs name for galaxy truth (e.g. cosmoDC2)
        config_path     Where to write config file. Default is data
                        directory.
        catalog_name    If a config file is written this value is saved
                        there and is also part of the filename of the
                        config file.
        output_type     A format.  For now only parquet is supported
        mag_cut         If not None, exclude galaxies with mag_r > mag_cut
        sed_subdir      In instcat entry, prepend this value to SED filename
        knots_mag_cut   No knots for galaxies with i_mag > cut
        knots           If True include knots
        logname         logname for Python logger
        pkg_root        defaults to three levels up from __file__
        skip_done       If True, skip over files which already exist. Otherwise
                        (by default) overwrite with new version.
                        Output info message in either case if file exists.
        flux_only       Only create flux files, not main files
        main_only       Only create main files, not flux files
        flux_parallel   Number of processes to divide work of computing fluxes

        Might want to add a way to specify template for output file name
        and template for input sedLookup file name.
        """

        _cosmo_cat = 'cosmodc2_v1.1.4_image_addon_knots'
        self._galaxy_stride = 1000000
        if pkg_root:
            self._pkg_root = pkg_root
        else:
            self._pkg_root = os.path.join(os.path.dirname(__file__),
                                          '../../..')

        if area_partition['type'] != 'healpix':
            raise NotImplementedError(f'CatalogCreator: Unknown partition type {area_partition["type"]} ')

        if output_type != 'parquet':
            raise NotImplementedError(f'CatalogCreator: Output type {output_type} not supported')

        self._galaxy_truth = galaxy_truth
        if galaxy_truth is None:
            self._galaxy_truth = _cosmo_cat

        self._sn_truth = None
        self._star_truth = None
        self._cat = None

        self._parts = parts
        self._area_partition = area_partition
        if skycatalog_root:
            self._skycatalog_root = skycatalog_root
        else:
            self._skycatalog_root = './'
        if catalog_dir:
            self._catalog_dir = catalog_dir
        else:
            self._catalog_dir = '.'

        self._output_dir = os.path.join(self._skycatalog_root,
                                        self._catalog_dir)

        self._written_config = None
        self._config_path = config_path
        self._catalog_name = catalog_name

        self._output_type = output_type
        self._mag_cut = mag_cut
        self._sed_subdir = sed_subdir
        self._knots_mag_cut = knots_mag_cut
        self._knots = knots
        self._logname = logname
        self._logger = logging.getLogger(logname)
        self._skip_done = skip_done
        self._flux_only = flux_only
        self._main_only = main_only
        self._flux_parallel = flux_parallel
        self._provenance = provenance

        self._obs_sed_factory = None

    def create(self, catalog_type):
        """
        Create catalog of specified type, using stored context.

        Parameters
        ----------
        catalog_type   string    Currently 'galaxy' and 'pointsource' are
                                 the only values allowed
        Return
        ------
        None
        """
        if catalog_type == ('galaxy'):
            if not self._flux_only:
                self.create_galaxy_catalog()
            if not self._main_only:
                self.create_galaxy_flux_catalog()
        elif catalog_type == ('pointsource'):
            if not self._flux_only:
                self.create_pointsource_catalog()
            if not self._main_only:
                self.create_pointsource_flux_catalog()
        else:
            raise NotImplemented(f'CatalogCreator.create: unsupported catalog type {catalog_type}')

    def create_galaxy_catalog(self):
        """
        Create the 'main' galaxy catalog, including everything except
        LSST fluxes

        Returns
        -------
        None

        """
        import GCRCatalogs

        self._cat = None

        gal_cat = GCRCatalogs.load_catalog(self._galaxy_truth)

        # Save cosmology in case we need to write parameters out later
        self._cosmology = gal_cat.cosmology

        ###self._mag_norm_f = MagNorm(self._cosmology)

        arrow_schema = make_galaxy_schema(self._logname,
                                              self._sed_subdir,
                                              self._knots)

        for p in self._parts:
            self._logger.info(f'Starting on pixel {p}')
            self.create_galaxy_pixel(p, gal_cat, arrow_schema)
            self._logger.info(f'Completed pixel {p}')

        # Now make config.   We need it for computing LSST fluxes for
        # the second part of the galaxy catalog
        if self._skip_done:
            config_path = self.write_config(path_only=True)
            if os.path.exists(config_path):
                self._logger.info('Will not overwrite existing config file')
                return
        self.write_config()

    def create_galaxy_pixel(self, pixel, gal_cat, arrow_schema):
        """
        Parameters
        ----------
        pixel           Pixel for which catalog is to be generated.
        gal_cat         GCRCatalogs-loaded galaxy truth (e.g. cosmoDC2)
        arrow_schema    schema to use for output file
        """

        output_path = os.path.join(self._output_dir, f'galaxy_{pixel}.parquet')
        if os.path.exists(output_path):
            if not self._skip_done:
                self._logger.info(f'Overwriting file {output_path}')
            else:
                self._logger.info(f'Skipping regeneration of {output_path}')
                return

        # Used to finding tophat parameters from cosmoDC2 column names
        tophat_bulge_re = r'sed_(?P<start>\d+)_(?P<width>\d+)_bulge'
        tophat_disk_re = r'sed_(?P<start>\d+)_(?P<width>\d+)_disk'

        # Number of rows to include in a row group
        #stride = 1000000
        stride = self._galaxy_stride

        hp_filter = [f'healpix_pixel=={pixel}']
        if self._mag_cut:
            r_mag_name = 'mag_r_lsst'
            mag_cut_filter = [f'{r_mag_name} < {self._mag_cut}']

        # to_fetch = all columns of interest in gal_cat
        non_sed = ['galaxy_id', 'ra', 'dec', 'redshift', 'redshiftHubble',
                   'peculiarVelocity', 'shear_1', 'shear_2',
                   'convergence', 'position_angle_true',
                   'size_bulge_true', 'size_minor_bulge_true', 'sersic_bulge',
                   'size_disk_true', 'size_minor_disk_true', 'sersic_disk']
        if self._knots:
                   non_sed += ['knots_flux_ratio', 'n_knots', 'mag_i_lsst']
        # Find all the tophat sed numbers
        q = gal_cat.list_all_quantities()
        sed_bulge_names = [i for i in q if (i.startswith('sed') and
                                            i.endswith('bulge'))]
        sed_disk_names = [i for i in q if (i.startswith('sed') and
                                           i.endswith('disk'))]
        # Save all the start, width values
        self._sed_bins = [[int(re.match(tophat_bulge_re, s)['start']), int(re.match(tophat_bulge_re, s)['width'])] for s in sed_bulge_names]

        # Sort by value for start
        def _bin_start_key(start_width):
            return start_width[0]
        self._sed_bins.sort(key=_bin_start_key)

        self._obs_sed_factory = ObservedSedFactory(self._sed_bins,
                                                   assemble_cosmology(self._cosmology))

        def _sed_bulge_key(s):
            return int(re.match(tophat_bulge_re, s)['start'])
        def _sed_disk_key(s):
            return int(re.match(tophat_disk_re, s)['start'])

        # Sort into increaing order by start wavelength
        sed_bulge_names.sort(key=_sed_bulge_key)
        sed_disk_names.sort(key=_sed_disk_key)
        if self._knots:
            sed_knot_names = [i.replace('disk', 'knots') for i in sed_disk_names]

        #Fetch the data
        to_fetch = non_sed + sed_bulge_names + sed_disk_names

        if not self._mag_cut:
            df = gal_cat.get_quantities(to_fetch, native_filters=hp_filter)
        else:
            df = gal_cat.get_quantities(to_fetch + [r_mag_name],
                                        native_filters=hp_filter,
                                        filters=mag_cut_filter)

        if self._sed_subdir:
            #  Generate full paths for disk and bulge SED files, even though
            #  we don't actually write the files here
            bulge_path = _generate_sed_path(df['galaxy_id'], self._sed_subdir,
                                            'bulge')
            disk_path =  _generate_sed_path(df['galaxy_id'], self._sed_subdir,
                                            'disk')

        if self._knots:
            # adjust disk sed; create knots sed
            eps = np.finfo(np.float32).eps
            mag_mask = np.where(np.array(df['mag_i_lsst']) > self._knots_mag_cut,0, 1)
            self._logger.debug(f'Count of mags <=  cut (so adjustment performed: {np.count_nonzero(mag_mask)}')

            for d_name, k_name in zip(sed_disk_names, sed_knot_names):
                df[k_name] = mag_mask * np.clip(df['knots_flux_ratio'], None, 1-eps) * df[d_name]
                df[d_name] = np.where(np.array(df['mag_i_lsst']) > self._knots_mag_cut, 1,
                                      np.clip(1 - df['knots_flux_ratio'], eps, None)) * df[d_name]
        # Re-form sed columns into two arrays
        bulge_seds = (np.array([df[sbn] for sbn in sed_bulge_names]).T).tolist()
        disk_seds = (np.array([df[sdn] for sdn in sed_disk_names]).T).tolist()
        if self._knots:
            knots_seds = (np.array([df[kdn] for kdn in sed_knot_names]).T).tolist()

        #  Compute mag_norm from TH sed and redshift
        bulge_magnorm = [self._obs_sed_factory.magnorm(s, z) for (s, z)\
                         in zip(bulge_seds, df['redshiftHubble']) ]
        disk_magnorm = [self._obs_sed_factory.magnorm(s, z) for (s, z)\
                        in zip(disk_seds, df['redshiftHubble']) ]
        if self._knots:
            knots_magnorm = [self._obs_sed_factory.magnorm(s, z) for (s, z)\
                             in zip(knots_seds, df['redshiftHubble']) ]

        MW_rv = np.full_like(df['sersic_bulge'], _MW_rv_constant)
        MW_av = _make_MW_extinction(df['ra'], df['dec'])
        self._logger.debug('Made extinction')

        #  Write row groups of size stride (or less) until input is exhausted
        last_row_ix = len(df['galaxy_id']) - 1
        u_bnd = min(stride, last_row_ix + 1)
        l_bnd = 0
        rg_written = 0
        writer = None

        # Some columns need to be renamed
        to_modify = ['position_angle_true', 'redshiftHubble', 'peculiarVelocity']
        while u_bnd > l_bnd:
            out_dict = {k : df[k][l_bnd : u_bnd] for k in non_sed if k not in to_modify}
            out_dict['redshift_hubble'] = df['redshiftHubble'][l_bnd : u_bnd]
            out_dict['peculiar_velocity'] = df['peculiarVelocity'][l_bnd : u_bnd]
            out_dict['position_angle_unlensed'] = df['position_angle_true'][l_bnd : u_bnd]
            out_dict['sed_val_bulge'] = bulge_seds[l_bnd : u_bnd]
            out_dict['sed_val_disk'] = disk_seds[l_bnd : u_bnd]
            out_dict['bulge_magnorm'] = bulge_magnorm[l_bnd : u_bnd]
            out_dict['disk_magnorm'] = disk_magnorm[l_bnd : u_bnd]
            out_dict['MW_rv'] = MW_rv[l_bnd : u_bnd]
            out_dict['MW_av'] = MW_av[l_bnd : u_bnd]

            if self._knots:
                out_dict['sed_val_knots'] = knots_seds[l_bnd : u_bnd]
                out_dict['n_knots'] = df['n_knots'][l_bnd : u_bnd]
                out_dict['knots_magnorm'] = knots_magnorm[l_bnd : u_bnd]

            if self._sed_subdir:
                out_dict['bulge_sed_file_path'] = bulge_path[l_bnd : u_bnd]
                out_dict['disk_sed_file_path'] = disk_path[l_bnd : u_bnd]

            out_df = pd.DataFrame.from_dict(out_dict)
            out_table = pa.Table.from_pandas(out_df, schema=arrow_schema)
            if not writer:
                writer = pq.ParquetWriter(output_path, arrow_schema)

            writer.write_table(out_table)
            rg_written += 1
            l_bnd = u_bnd
            u_bnd = min(l_bnd + stride, last_row_ix + 1)

        writer.close()
        self._logger.debug(f'# row groups written: {rg_written}')
        if self._provenance == 'yaml':
            self.write_provenance_file(output_path)

    def create_galaxy_flux_catalog(self, config_file=None):
        '''
        Create a second file per healpixel containing just galaxy id and
        LSST fluxes.  Use information in the main file to compute fluxes

        Parameters
        ----------
        Path to config created in first stage so we can find the main
        galaxy files.

        Return
        ------
        None
        '''

        from desc.skycatalogs import open_catalog, SkyCatalog

        self._gal_flux_schema = make_galaxy_flux_schema(self._logname)

        if not config_file:
            #config_file = self._written_config
            config_file = self.write_config(path_only=True)
        if not self._cat:
            self._cat = open_catalog(config_file,
                                     skycatalog_root=self._skycatalog_root)

        # Might also open a main file and read its metadata to be sure
        # the value for stride is correct.
        # Especially need to do this if we want to support making flux
        # files in a separate job activation.

        self._flux_template = self._cat.raw_config['object_types']['galaxy']['flux_file_template']

        self._logger.info('Creating galaxy flux files')
        for p in self._parts:
            self._logger.info(f'Starting on pixel {p}')
            self._create_galaxy_flux_pixel(p)
            self._logger.info(f'Completed pixel {p}')


    def _create_galaxy_flux_pixel(self, pixel):
        '''
        Create a parquet file for a single healpix pixel containing only
        galaxy id and LSST fluxes

        Parameters
        ----------
        Pixel         int

        Return
        ------
        None
        '''

        output_filename = f'galaxy_flux_{pixel}.parquet'
        output_path = os.path.join(self._output_dir, output_filename)

        if os.path.exists(output_path):
            if not self._skip_done:
                self._logger.info(f'Overwriting file {output_path}')
            else:
                self._logger.info(f'Skipping regeneration of {output_path}')
                return

        # Use same value for row group as was used for main file
        stride = self._galaxy_stride

        object_list = self._cat.get_objects_by_hp(pixel,
                                                  obj_type_set={'galaxy'})
        object_coll = object_list.get_collections()[0]
        writer = None
        l_bnd = 0
        u_bnd = min(len(object_coll), stride)
        rg_written = 0
        while u_bnd > l_bnd:
            o_list = object_coll[l_bnd : u_bnd]
            self._logger.debug(f'Handling range {l_bnd} up to {u_bnd}')

            # prefetch everything we need. Getting a quantity for one object
            # ensures the quantity is read for the whole row group
            # (In fact currently all row groups are read, but that could and
            # should change)
            for att in ['galaxy_id', 'shear_1', 'shear_2', 'convergence',
                        'redshift_hubble', 'MW_av', 'MW_rv', 'sed_val_bulge',
                        'sed_val_disk', 'sed_val_knots'] :
                v = o_list[0].get_native_attribute(att)

            global _galaxy_collection
            _galaxy_collection = object_coll

            out_dict = {'galaxy_id': [], 'lsst_flux_u' : [],
                        'lsst_flux_g' : [], 'lsst_flux_r' : [],
                        'lsst_flux_i' : [], 'lsst_flux_z' : [],
                        'lsst_flux_y' : []}

            n_parallel = self._flux_parallel

            if n_parallel == 1:
                n_per = u_bnd - l_bnd
            else:
                n_per = int((u_bnd - l_bnd + n_parallel)/n_parallel)
            l = l_bnd
            u = min(l_bnd + n_per, u_bnd)
            readers = []

            if n_parallel == 1:
                out_dict = _do_galaxy_flux_chunk(None, _galaxy_collection,
                                                 l, u)
            else:
                # Expect to be able to do about 1500/minute/process
                tm = int((n_per*60)/500)  # Give ourselves a cushion
                self._logger.info(f'Using timeout value {tm} for {n_per} sources')
                p_list = []
                for i in range(n_parallel):
                    conn_rd, conn_wrt = Pipe(duplex=False)
                    readers.append(conn_rd)

                    # For debugging call directly
                    proc = Process(target=_do_galaxy_flux_chunk,
                                   name=f'proc_{i}',
                                   args=(conn_wrt, _galaxy_collection,l, u))
                    proc.start()
                    p_list.append(proc)
                    l = u
                    u = min(l + n_per, u_bnd)

                self._logger.debug('Processes started')

                for i in range(n_parallel):
                    ready = readers[i].poll(tm)
                    if not ready:
                        self._logger.error(f'Process {i} timed out after {tm} sec')
                        sys.exit(1)
                    dat = readers[i].recv()
                    for k in ['galaxy_id', 'lsst_flux_u', 'lsst_flux_g',
                              'lsst_flux_r', 'lsst_flux_i', 'lsst_flux_z',
                              'lsst_flux_y']:
                        out_dict[k] += dat[k]
                for p in p_list:
                    p.join()

            out_df = pd.DataFrame.from_dict(out_dict)
            out_table = pa.Table.from_pandas(out_df,
                                             schema=self._gal_flux_schema)

            if not writer:
                writer = pq.ParquetWriter(output_path, self._gal_flux_schema)
            writer.write_table(out_table)
            l_bnd = u_bnd
            u_bnd = min(l_bnd + stride, len(object_coll))

            rg_written +=1

        writer.close()
        self._logger.debug(f'# row groups written to flux file: {rg_written}')
        if self._provenance == 'yaml':
            self.write_provenance_file(output_path)

    def create_pointsource_catalog(self, star_truth=None, sn_truth=None):

        """
        Parameters
        ----------
        star_truth      Where to find star parameters. If None use default
        sn_truth       Where to find SN parameters.  If None, use default

        Might want to add a way to specify template for output file name

        Returns
        -------
        None
        """

        # For now fixed location for star, SNe parameter files.
        _star_db = '/global/cfs/cdirs/lsst/groups/SSim/DC2/dc2_stellar_healpixel.db'
        _sn_db = '/global/cfs/cdirs/lsst/groups/SSim/DC2/cosmoDC2_v1.1.4/sne_cosmoDC2_v1.1.4_MS_DDF_healpix.db'
        self._star_truth = _star_db
        self._sn_truth = _sn_db

        arrow_schema = make_star_schema()
        #  Need a way to indicate which object types to include; deal with that
        #  later.  For now, default is stars only.  Use default star parameter file.
        for p in self._parts:
            self._logger.debug(f'Point sources. Starting on pixel {p}')
            self.create_pointsource_pixel(p, arrow_schema, star_cat=_star_db)
            self._logger.debug(f'Completed pixel {p}')

    def create_pointsource_pixel(self, pixel, arrow_schema, star_cat=None,
                             sn_cat=None):
        if not star_cat and not sn_cat:
            self._logger.info('No point source inputs specified')
            return

        output_filename = f'pointsource_{pixel}.parquet'
        output_path = os.path.join(self._output_dir, output_filename)

        if os.path.exists(output_path):
            if not self._skip_done:
                self._logger.info(f'Overwriting file {output_path}')
            else:
                self._logger.info(f'Skipping regeneration of {output_path}')
                return

        if star_cat:
            # Get data for this pixel
            cols = ','.join(['simobjid as id', 'ra', 'decl as dec',
                             'magNorm as magnorm', 'sedFilename as sed_filepath'])
            q = f'select {cols} from stars where hpid={pixel} '
            with sqlite3.connect(star_cat) as conn:
                star_df = pd.read_sql_query(q, conn)

            star_df['sed_filepath'] = get_star_sed_path(star_df['sed_filepath'])

            nobj = len(star_df['id'])
            self._logger.debug(f'Found {nobj} stars')
            star_df['object_type'] = np.full((nobj,), 'star')
            star_df['host_galaxy_id'] = np.zeros((nobj,), np.int64())

            star_df['MW_rv'] = np.full((nobj,), _MW_rv_constant, np.float32())

            star_df['MW_av'] = _make_MW_extinction(np.array(star_df['ra']),
                                                   np.array(star_df['dec']))
            out_table = pa.Table.from_pandas(star_df, schema=arrow_schema)
            self._logger.debug('Created arrow table from dataframe')

            writer = pq.ParquetWriter(output_path, arrow_schema)
            writer.write_table(out_table)

            writer.close()
            if self._provenance == 'yaml':
                self.write_provenance_file(output_path)

        if sn_cat:
            raise NotImplementedError('SNe not yet supported. Have a nice day.')

        return

    def create_pointsource_flux_catalog(self, config_file=None):
        '''
        Create a second file per healpixel containing just id and
        LSST fluxes.  Use information in the main file to compute fluxes

        Parameters
        ----------
        Path to config created in first stage so we can find the main
        galaxy files.

        Return
        ------
        None
        '''

        from desc.skycatalogs import open_catalog, SkyCatalog

        self._ps_flux_schema = make_star_flux_schema(self._logname)
        if not config_file:
            config_file = self.write_config(path_only=True)

        # Always open catalog. If it was opened for galaxies earlier
        # it won't know about star files.
        self._cat = open_catalog(config_file,
                                 skycatalog_root=self._skycatalog_root)

        self._flux_template = self._cat.raw_config['object_types']['star']['flux_file_template']

        self._logger.info('Creating pointsource flux files')
        for p in self._parts:
            self._logger.info(f'Starting on pixel {p}')
            self._create_pointsource_flux_pixel(p)
            self._logger.info(f'Completed pixel {p}')

    def _create_pointsource_flux_pixel(self, pixel):
        '''
        Create a parquet file for a single healpix pixel containing only
        pointsource id and LSST fluxes

        Parameters
        ----------
        pixel         int

        Return
        ------
        None
        '''

        # For main catalog use self._cat
        # For schema use self._ps_flux_schema
        # output_template should be derived from value for flux_file_template
        #  in main catalog config.  Cheat for now
        output_filename = f'pointsource_flux_{pixel}.parquet'
        output_path = os.path.join(self._output_dir, output_filename)

        if os.path.exists(output_path):
            if not self._skip_done:
                self._logger.info(f'Overwriting file {output_path}')
            else:
                self._logger.info(f'Skipping regeneration of {output_path}')
                return

        object_list = self._cat.get_objects_by_hp(pixel,
                                                  obj_type_set={'star'})
        last_row_ix = len(object_list) - 1
        writer = None

        # Write out as a single rowgroup as was done for main catalog
        l_bnd = 0
        u_bnd = last_row_ix + 1

        o_list = object_list[l_bnd : u_bnd]
        self._logger.debug(f'Handling range {l_bnd} up to {u_bnd}')
        out_dict = {}
        out_dict['id'] = [o.get_native_attribute('id') for o in o_list]
        all_fluxes =  [o.get_LSST_fluxes(as_dict=False) for o in o_list]
        all_fluxes_transpose = zip(*all_fluxes)
        for i, band in enumerate(LSST_BANDS):
            self._logger.debug(f'Band {band} is number {i}')
            v = all_fluxes_transpose.__next__()
            out_dict[f'lsst_flux_{band}'] = v
            if i == 1:
                self._logger.debug(f'Len of flux column: {len(v)}')
                self._logger.debug(f'Type of flux column: {type(v)}')
        out_df = pd.DataFrame.from_dict(out_dict)
        out_table = pa.Table.from_pandas(out_df,
                                         schema=self._ps_flux_schema)
        if not writer:
            writer = pq.ParquetWriter(output_path, self._ps_flux_schema)
        writer.write_table(out_table)

        writer.close()
        ##self._logger.debug(f'# row groups written to flux file: {rg_written}')
        if self._provenance == 'yaml':
            self.write_provenance_file(output_path)

    def write_config(self, overwrite=False, path_only=False):
        '''
        Parameters
        ----------
        overwrite   boolean default False.   If true, overwrite existing
                    config of the same name
        path_only   If true, just return the path; don't write anything

        Returns
        -------
        Path to would-be config file if path_only is True;
        else None

        Side-effects
        ------------
        Save path to config file written as instance variable

        '''
        if not self._config_path:
            self._config_path = self._output_dir

        if path_only:
            return os.path.join(self._config_path, self._catalog_name + '.yaml')


        config = create_config(self._catalog_name, self._logname)
        config.add_key('area_partition', self._area_partition)
        config.add_key('skycatalog_root', self._skycatalog_root)
        config.add_key('catalog_dir' , self._catalog_dir)

        config.add_key('SED_models',
                       assemble_SED_models(self._sed_bins))
        config.add_key('MW_extinction_values', assemble_MW_extinction())
        config.add_key('Cosmology', assemble_cosmology(self._cosmology))
        config.add_key('object_types', assemble_object_types(self._pkg_root))

        config.add_key('galaxy_magnitude_cut', self._mag_cut)
        config.add_key('knots_magnitude_cut', self._knots_mag_cut)

        inputs = {'galaxy_truth' : self._galaxy_truth}
        if self._sn_truth:
            inputs['sn_truth'] = self._sn_truth
        if self._star_truth:
            inputs['star_truth'] = self._star_truth
        config.add_key('provenance', assemble_provenance(self._pkg_root,
                                                         inputs=inputs))

        self._written_config = config.write_config(self._config_path,
                                                   overwrite=overwrite)

    def write_provenance_file(self, datafile_path):
        '''
        Write git provenance to a yaml file with name derived from a
        just-written datafile name
        '''
        outpath = datafile_path.rsplit('.', 1)[0] + '_provenance.yaml'

        prov = assemble_provenance(self._pkg_root, inputs=None)
        write_yaml(prov, outpath)



from collections import namedtuple, OrderedDict
from enum import Enum
import numpy as np

__all__ = ['column_finder', 'check_file', 'write_to_instance', 'SourceType',
           'STAR_FMT', 'CMP_FMT',
           'form_star_instance_columns', 'form_cmp_instance_columns']


STAR_FMT = '{:s} {:d} {:.14f} {:.14f} {:.8f} {:s} {:d} {:d} {:d} {:d} {:d} {:d} {:s} {:s} {:s} {:.8f} {:f}'

CMP_FMT = '{:s} {:s} {:.14f} {:.14f} {:.8f} {:s} {:.9g} {:.9g} {:.9g} {:.9g} {:d} {:d} {:s} {:.9g} {:.9g} {:f} {:.0f} {:s} {:s} {:.8f} {:f}'

def form_star_instance_columns(band):
    star_instance = [column_finder('prefix', SourceType.FIXED, ('object', np.dtype('U6'))),
                     column_finder('uniqueId', SourceType.DATA, 'id'),
                     column_finder('raPhoSim', SourceType.DATA, 'ra'),
                     column_finder('decPhoSim', SourceType.DATA, 'dec'),
                     column_finder('maskedMagNorm', SourceType.DATA, 'magnorm'),
                     column_finder('sedFilepath',SourceType.DATA, 'sed_filepath'),
                     column_finder('redshift', SourceType.FIXED, (0, int)),
                     column_finder('gamma1', SourceType.FIXED, (0, int)),
                     column_finder('gamma2', SourceType.FIXED, (0, int)),
                     column_finder('kappa', SourceType.FIXED, (0, int)),
                     column_finder('raOffset', SourceType.FIXED, (0, int)),
                     column_finder('decOffset', SourceType.FIXED, (0, int)),
                     column_finder('spatialmodel', SourceType.FIXED, ('point', np.dtype('U5'))),
                     column_finder('internalExtinctionModel', SourceType.FIXED, ('none', np.dtype('U4'))),
                     column_finder('galacticExtinctionModel', SourceType.CONFIG,
                                   ('object_types/star/MW_extinction', str)),
                     column_finder('galactivAv', SourceType.DATA,
                                   'MW_av'),
                     column_finder('galacticRv', SourceType.CONFIG,
                                   ('MW_extinction_values/r_v/value', float))]
    return star_instance

def _form_knots_instance_columns(cmp, band):
    cmp_instance = [column_finder('prefix', SourceType.FIXED, ('object', np.dtype('U6'))),
                    column_finder('uniqueId', SourceType.COMPUTE, ['galaxy_id', f'{cmp}']),
                    column_finder('raPhoSim', SourceType.DATA, 'ra'),
                    column_finder('decPhoSim', SourceType.DATA, 'dec'),
                    column_finder('phoSimMagNorm', SourceType.DATA, 'knots_magnorm'),
                    column_finder('sedFilepath',SourceType.COMPUTE, [f'sed_val_{cmp}','redshift_hubble']),
                    column_finder('redshift', SourceType.DATA, 'redshift'),
                    column_finder('gamma1', SourceType.DATA, 'shear_1'),
                    column_finder('gamma2', SourceType.DATA, 'shear_2'),
                    column_finder('kappa', SourceType.DATA, 'convergence'),
                    column_finder('raOffset', SourceType.FIXED, (0, int)),
                    column_finder('decOffset', SourceType.FIXED, (0, int)),
                    column_finder('spatialmodel', SourceType.CONFIG,
                                  (f'object_types/{cmp}_basic/spatial_model', 'str')),
                    column_finder('majorAxis', SourceType.DATA, 'size_disk_true'),
                    column_finder('minorAxis', SourceType.DATA, 'size_minor_disk_true'),
                    column_finder('positionAngle', SourceType.DATA, 'position_angle_unlensed'),
                    column_finder('sindex', SourceType.DATA, 'n_knots'),
                    column_finder('internalExtinctionModel', SourceType.FIXED, ('none', np.dtype('U4'))),
                    column_finder('galacticExtinctionModel', SourceType.CONFIG,
                                  (f'object_types/{cmp}_basic/MW_extinction', 'str')),
                    column_finder('galactivAv', SourceType.DATA,
                                  'MW_av'),
                    column_finder('galacticRv', SourceType.CONFIG,
                                  ('MW_extinction_values/r_v/value', float))]
    return cmp_instance

###
def form_cmp_instance_columns(cmp, band):
    if cmp == 'knots':
        return _form_knots_instance_columns(cmp, band)

    cmp_instance = [column_finder('prefix', SourceType.FIXED, ('object', np.dtype('U6'))),
                    column_finder('uniqueId', SourceType.COMPUTE, ['galaxy_id', f'{cmp}']),
                    column_finder('raPhoSim', SourceType.DATA, 'ra'),
                    column_finder('decPhoSim', SourceType.DATA, 'dec'),
                    column_finder('phoSimMagNorm', SourceType.DATA, f'{cmp}_magnorm'),
                    column_finder('sedFilepath',SourceType.COMPUTE, [f'sed_val_{cmp}','redshift_hubble']),
                    column_finder('redshift', SourceType.DATA, 'redshift'),
                    column_finder('gamma1', SourceType.DATA, 'shear_1'),
                    column_finder('gamma2', SourceType.DATA, 'shear_2'),
                    column_finder('kappa', SourceType.DATA, 'convergence'),
                    column_finder('raOffset', SourceType.FIXED, (0, int)),
                    column_finder('decOffset', SourceType.FIXED, (0, int)),
                    column_finder('spatialmodel', SourceType.CONFIG,
                                  (f'object_types/{cmp}_basic/spatial_model', 'str')),
                    column_finder('majorAxis', SourceType.DATA, f'size_{cmp}_true'),
                    column_finder('minorAxis', SourceType.DATA, f'size_minor_{cmp}_true'),
                    column_finder('positionAngle', SourceType.DATA, 'position_angle_unlensed'),
                    column_finder('sindex', SourceType.DATA, f'sersic_{cmp}'),
                    column_finder('internalExtinctionModel', SourceType.FIXED, ('none', np.dtype('U4'))),
                    column_finder('galacticExtinctionModel', SourceType.CONFIG,
                                  (f'object_types/{cmp}_basic/MW_extinction', 'str')),
                    column_finder('galactivAv', SourceType.DATA,
                                  'MW_av'),
                    column_finder('galacticRv', SourceType.CONFIG,
                                  ('MW_extinction_values/r_v/value', float))]
    return cmp_instance

###
column_finder = namedtuple('ColumnFinder', ['instance_name', 'source_type',
                                            'source_parm'])
SourceType = Enum('SourceType', 'DATA CONFIG FIXED COMPUTE')
'''
Used in source_type field of column_finder to describe source of each
instance catalog value

DATA    Sky catalog column
CONFIG  Sky catalog config field
FIXED   Fixed value supplied inline
COMPUTE Arbitrary computation which may involve any of the above
'''

def check_file(path):
    '''Look for a file that should not exist'''
    try:
        f = open(path, mode='r')
    except FileNotFoundError as e:
        return

    raise ValueError(f'File for {path} already exists')

def write_to_instance(handle, ordered_data_dict, fmt):
    '''
    Parameters
    ----------
    handle             file handle to which text will be written
    ordered_data_dict  columns in the order they should be written
    '''
    col_list = [ordered_data_dict[k] for k in ordered_data_dict]
    for row in zip(*col_list):
        handle.write(fmt.format(*row) + '\n')

def form_object_string(obj, band, component):
    '''
    parse columns for this object/component type
    fetch data and config values as needed
    form values into a list.
    '''
    row = []

    if obj.object_type == 'star':
        fmt = STAR_FMT
        columns = form_star_instance_columns(band)
    elif obj.object_type in ['galaxy', 'disk', 'bulge']:
        fmt = CMP_FMT
        if obj.object_type == 'galaxy':
            cmp = component
        else:
            cmp = obj.object_type
        columns = form_cmp_instance_columns(cmp, band)
    else:
        raise NotImplementedError(f'translate_utils.form_object_string: Unsupported object type {obj.object_type}')

    for c in columns:
        if c.source_type == SourceType.DATA:
            row.append(obj.get_native_attribute(c.source_parm))
        elif c.source_type == SourceType.CONFIG:
            v = obj.belongs_to.config.get_config_value(c.source_parm[0])
            t = c.source_parm[1]
            if str(t) in ['float', 'int']:
                q = eval(f'{t}({v})')
            else:
                q = v
            row.append(q)
        elif c.source_type == SourceType.FIXED:
            v = c.source_parm[0]
            t = c.source_parm[1]
            if str(t) in ['float', 'int']:
                q = eval(f'{t}({v})')
            else:
                q = v
            row.append(q)
        elif c.source_type == SourceType.COMPUTE:
            # only one is sedFilepath, and only for galaxy components
            if c.instance_name not in ['sedFilepath', 'uniqueId'] or cmp not in ['disk', 'bulge', 'knots']:
                raise ValueError(f'translate_utils.form_object_string: Bad COMPUTE entry {c.instance_name} for component {cmp}')
            if c.instance_name == 'sedFilepath':
                row.append(f'{obj.get_native_attribute("galaxy_id")}_{cmp}_sed.txt')
            else:     # uniqueId
                row.append(f'{str(obj.get_native_attribute("galaxy_id"))}_{cmp}')
        else:
            raise ValueError(f'translate_utils.form_object_string: Unknown source type {c.source_type}')

    return write_to_string(row, fmt)

def write_to_string(row, fmt):
    '''
    Output single string, composed from input values in row.
    Types of values in row must match those expected by fmt.
    Parameters
    ----------
    row     list of values
    fmt     fmt string.

    Returns
    -------
    Formatted string

    '''
    return(fmt.format(*row))



import pyarrow as pa
import logging

__all__ = ['make_galaxy_schema', 'make_galaxy_flux_schema', 'make_star_schema']

# This schema is not the same as the one taken from the data,
# probably because of the indexing in the schema derived from a pandas df.
def make_galaxy_schema(logname, sed_subdir=False, knots=True):
    fields = [pa.field('galaxy_id', pa.int64()),
              pa.field('ra', pa.float64() , True),
##                       metadata={"units" : "radians"}),
              pa.field('dec', pa.float64() , True),
##                       metadata={"units" : "radians"}),
              pa.field('redshift', pa.float64(), True),
              pa.field('redshift_hubble', pa.float64(), True),
              pa.field('peculiar_velocity', pa.float64(), True),
              pa.field('shear_1', pa.float64(), True),
              pa.field('shear_2', pa.float64(), True),
              pa.field('convergence', pa.float64(), True),
              pa.field('size_bulge_true', pa.float32(), True),
              pa.field('size_minor_bulge_true', pa.float32(), True),
              pa.field('sersic_bulge', pa.float32(), True),
              pa.field('size_disk_true', pa.float32(), True),
              pa.field('size_minor_disk_true', pa.float32(), True),
              pa.field('sersic_disk', pa.float32(), True),
              pa.field('position_angle_unlensed', pa.float64(), True),
              pa.field('sed_val_bulge',
                       pa.list_(pa.float64()), True),
              pa.field('sed_val_disk',
                       pa.list_(pa.float64()), True),
              pa.field('bulge_magnorm', pa.float64(), True),
              pa.field('disk_magnorm', pa.float64(), True),
              pa.field('MW_rv', pa.float32(), True),
              pa.field('MW_av', pa.float32(), True)]
    logger = logging.getLogger(logname)
    if knots:
        logger.debug("knots requested")
        fields.append(pa.field('sed_val_knots',
                               pa.list_(pa.float64()), True))
        ### For sizes API can alias to disk sizes
        ###  position angle, shears and convergence are all
        ###  galaxy-wide quantities.
        fields.append(pa.field('n_knots', pa.float32(), True))
        fields.append(pa.field('knots_magnorm', pa.float64(), True))

    if sed_subdir:
        fields.append(pa.field('bulge_sed_file_path', pa.string(), True))
        fields.append(pa.field('disk_sed_file_path', pa.string(), True))

    debug_out = ''
    for f in fields:
        debug_out += f'{f.name}\n'
    logger.debug(debug_out)
    return pa.schema(fields)

def make_galaxy_flux_schema(logname):
    '''
    Will make a separate parquet file with lsst flux for each band
    and galaxy id for joining with the main galaxy file
    '''
    logger = logging.getLogger(logname)
    logger.debug('Creating galaxy flux schema')
    fields = [pa.field('galaxy_id', pa.int64()),
              # should flux fields be named e.g. lsst_cmodel_flux_u?
              pa.field('lsst_flux_u', pa.float32() , True),
              pa.field('lsst_flux_g', pa.float32() , True),
              pa.field('lsst_flux_r', pa.float32() , True),
              pa.field('lsst_flux_i', pa.float32() , True),
              pa.field('lsst_flux_z', pa.float32() , True),
              pa.field('lsst_flux_y', pa.float32() , True)]
    return pa.schema(fields)

def make_star_schema():
    '''
    Minimal schema for non-variable stars.  For variables will need to add fields
    to express variability.   Will also likely have to make changes to accomodate SNe.
    If AGN also go in this file will need to include gamma1, gamma2, kappa.
    Could add field for galactic extinction model, but currently it's always 'CCM'
    so will put it in config.
    '''
    fields = [pa.field('object_type', pa.string(), False),
              pa.field('id', pa.int64(), False),
              pa.field('ra', pa.float64(), False),
              pa.field('dec', pa.float64(), False),
              pa.field('host_galaxy_id', pa.int64(), True),
              pa.field('magnorm', pa.float64(), True),
              pa.field('sed_filepath', pa.string(), True),
              pa.field('MW_rv', pa.float32(), True),
              pa.field('MW_av', pa.float32(), True)]
    return pa.schema(fields)

def make_star_flux_schema(logname):
    '''
    Will make a separate parquet file with lsst flux for each band
    and id for joining with the main star file
    '''
    logger = logging.getLogger(logname)
    logger.debug('Creating star flux schema')
    fields = [pa.field('id', pa.int64()),
              pa.field('lsst_flux_u', pa.float32() , True),
              pa.field('lsst_flux_g', pa.float32() , True),
              pa.field('lsst_flux_r', pa.float32() , True),
              pa.field('lsst_flux_i', pa.float32() , True),
              pa.field('lsst_flux_z', pa.float32() , True),
              pa.field('lsst_flux_y', pa.float32() , True)]
    return pa.schema(fields)



import sys
import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
import pandas as pd

def make_parquet(input_path):
    '''
    Given a text file in table format where columns are wavelength (nanometers)
    and flux, convert to parquet
    '''

    wv,flux = np.genfromtxt(input_path, unpack=True)
    #wv32 = np.array(wv, np.float32)

    df = pd.DataFrame({'wavelength': wv, 'flux': flux})
    #df = pd.DataFrame({'wavelength': wv32, 'flux': flux})
    # This method produces a file somewhat larger than if we left wv alone

    table = pa.Table.from_pandas(df)

    # as does this
    # table = pa.Table.from_arrays([wv32, flux], names=['wavelength', 'flux'])

    out_path = input_path + '.parquet'
    pq.write_table(table, out_path)

if __name__ == '__main__':
    if len(sys.argv) == 1:
        print('Requires filepath argument')
        exit(0)

    fname = sys.argv[1]

    print('Called with filepath argument ',fname)

    mode = 'parquet'

    if len(sys.argv) > 2:
        mode = sys.argv[2]

    if mode == 'parquet':
        make_parquet(fname)



__all__ = ['SkyCatalogsException', 'NoSchemaVersionError',
           'ConfigDuplicateKeyError']

class SkyCatalogsException(Exception):
    pass

class NoSchemaVersionError(SkyCatalogsException):
    def __init__(self, msg):

        if not msg:
            msg = 'Schema version unspecified'
        self.msg = msg
        super().__init__(self.msg)

class ConfigDuplicateKeyError(SkyCatalogsException):
    def __init__(self, key):
        self.msg = f'Cannot add duplicate key {key} to config'
        super().__init__(self.msg)



#  Copied from imSim repo
#
# Conventions:
#   M.m means a development version not ready for prime time yet.  Not stable
#   M.m.r is a regular release:
#      Intended to be API stable with previous M.x.y releases.
#      New features may be added with the major (M) series, but the intent is to not
#      break public API-based code from previous versions in the series.
#      Updates that change the revision number (r) should only have bug fixes.
#   M.m.r-* is some kind of prerelease or release candidate for alpha or beta testing.
#
# cf. https://semver.org/ for more info

# This should be updated before a release.
__version__ = '1.2.0'

# This will work if the version has end tags like 2.0.0-rc.1
__version_info__ = tuple(map(lambda x:int(x.split('-')[0]), __version__.split('.')))[:3]



import os
import re

from setuptools import setup, find_packages

with open('README.md') as file:
    long_description = file.read()

packages = find_packages(where="python")
print('packages = ', packages)

def all_files_from(dir, ext=''):
    """Quick function to get all files from directory and all subdirectories
    """
    files = []
    for root, dirnames, filenames in os.walk(dir):
        for filename in filenames:
            if filename.endswith(ext) and not filename.startswith('.'):
                files.append(os.path.join(root, filename))
    return files

# Need to clarify if this belongs in the distribution or not
## shared_data = all_files_from('data')
## print('shared_data = ', shared_data)

## configs = [os.path.join('cfg', 'latest.yaml')]

# Read in the version from python/desc/_version.py
# cf. http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package

version_file = os.path.join('python', 'desc', 'skycatalogs', '_version.py')
verstrline = open(version_file, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    skycatalogs_version = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (version_file,))
print('skyCatalogs version is %s'%(skycatalogs_version))

dist = setup(name="skyCatalogs",
             version=skycatalogs_version,
             maintainer="Joanne Bogart",                 # also author line?
             maintainer_email="jrb@slac.stanford.edu",
             license="BSD",
             description="Writes, reads catalogs input to LSST DESC simulations",
             long_description=long_description,
             package_dir={"": "python"},
             packages=find_packages(where="python"),
             ##package_data={"skycatalogs": shared_data + configs},
             url="https://github.com/LSSTDESC/skyCatalogs",
             classifiers=[
                 "License :: OSI Approved :: BSD License",
                 "Intended Audience :: Developers",
                 "Intended Audience :: Science/Research",
                 "Programming Language :: Python",
                 "Development Status :: 3 - Alpha",
                 ],
             install_requires=['numpy', 'healpy', 'astropy', 'pyarrow',
                               'pandas', 'galsim', 'dust_extinction']
             )



"""
Utilities for top-level scripts
"""
from datetime import datetime as dt
import sys
import logging

__all__ = ['print_callinfo', 'log_callinfo', 'print_date', 'print_dated_msg', 'TIME_TO_SECOND_FMT']

TIME_TO_SECOND_FMT = '%Y-%m-%d %H:%M:%S'

def print_callinfo(prog, args):
    """
    Print information about how a script using argparse was called

    Parameters
    ----------
    prog   program name, typically sys.argv[0]
    args   object returned by ArgumentParser.parse_args()
    """

    print('{}   {}  invoked  with arguments'.format(dt.now().strftime(TIME_TO_SECOND_FMT), prog))
    for e in dir(args):
        if not e.startswith('_'):
            nm = 'args.' + e
            print('{}: {}'.format(e, eval(nm)))

    sys.stdout.flush()

def log_callinfo(prog, args, logname):
    """
    Write information about how a script using argparse was called to log

    Parameters
    ----------
    prog   program name, typically sys.argv[0]
    args   object returned by ArgumentParser.parse_args()
    logname string
    """

    logger = logging.getLogger(logname)
    log_out = '{}  invoked  with arguments\n'.format(prog)
    for e in dir(args):
        if not e.startswith('_'):
            nm = 'args.' + e
            log_out += '    {}: {}\n'.format(e, eval(nm))
    logger.info(log_out)

def print_date(to_second=True, file=None):
    """
    Print current time (by default only to nearest second) and flush output
    Print to file if supplied, else to std.out
    """
    if to_second:
        print(dt.now().strftime(TIME_TO_SECOND_FMT), file=file, flush=True)
    else:
        print(dt.now(), file=file, flush=True)
def print_dated_msg(msg, to_second=True, file=None):
    """
    Print current time (by default only to nearest second) and flush output
    Print to file if supplied, else to std.out
    """
    if to_second:
        print(dt.now().strftime(TIME_TO_SECOND_FMT), ' ', msg,  file=file, flush=True)
    else:
        print(dt.now(), ' ', msg, file=file, flush=True)



import os
import yaml
import git
import logging
from jsonschema import validate

from desc.skycatalogs.utils.exceptions import NoSchemaVersionError, ConfigDuplicateKeyError

from collections import namedtuple

__all__ = ['Config', 'open_config_file', 'Tophat', 'create_config',
           'assemble_SED_models', 'assemble_MW_extinction',
           'assemble_cosmology', 'assemble_object_types', 'assemble_provenance',
           'write_yaml']

_CURRENT_SCHEMA_VERSION='1.1.0'

def open_config_file(config_file):
    '''
    Given path to config file, return a Config object
    '''
    with open(config_file) as f:
        return Config(yaml.safe_load(f))

Tophat = namedtuple('Tophat', ['start', 'width'])


def _custom_dir(c, add):
    '''
    dir should return
       * functions defined for c's class
       * instance variables belonging to c
       * public stuff from the contained object (some table-like thing)
    '''
    return dir(type(c)) + list(c.__dict__.keys()) + add


class DelegatorBase:
    '''
    This class allows derived classes to delegate operations to a
    member.   Here it is used to delegate dict operations to
    the member self._cfg
    '''

    @property
    def _delegate(self):
        pub = [o for o in dir(self.default) if not o.startswith('_')]
        return pub
    def __getattr__(self, k):
        if k in self._delegate:
            return getattr(self.default, k)
        raise AttributeError(k)
    def __dir__(self):
        return _custom_dir(self, self._delegate)

    def __init__(self):
        pass

class Config(DelegatorBase):
    '''
    A wrapper around the dict which is the contents of a Sky Catalog
    config file (a dict) which understands some of the semantics

    '''
    def __init__(self, cfg, logname=None):
        '''
        Parameters
        ----------
        cfg  is normally a dict, but it can also itself be a Config.
        In this case, store its dict to delegate to.
        '''
        if isinstance(cfg, Config):
            cfg = cfg._cfg

        self._cfg = cfg
        self._logname = logname

        super().__init__()

        self.default = cfg

    def __getitem__(self, k):
        '''
        Specific override of __getitem__, delegating to dict member
        so that [ ] syntax will be handled properly
        '''
        return self._cfg.__getitem__(k)

    def list_sed_models(self):
        return self._cfg['SED_models'].keys()

    def list_object_types(self):
        return self._cfg['object_types'].keys()

    def get_sed_model(self, modelname):
        return self._cfg['SED_models'][modelname]

    def object_is_composite(self, objectname):
        return 'composite' in self._cfg['object_types'][objectname]

    def get_object_parent(self, objectname):
        if 'parent' in self._cfg['object_types'][objectname]:
            return self._cfg['object_type'][objectname]['parent']
        else:
            return None

    def get_object_sedmodel(self, objectname):
        if 'sed_model' in self._cfg['object_types'][objectname]:
            return self._cfg['object_type'][objectname]['sed_model']
        else:
            return None

    def get_tophat_parameters(self):
        '''
        Return list of named tuples
        Should maybe be part of Sky Catalogs API
        '''
        raw_bins = self._cfg['SED_models']['tophat']['bins']

        return [ Tophat(b[0], b[1]) for b in raw_bins]

    def get_config_value(self, key_path):
        '''
        Find value belonging to key in a config.
        Parameters
        ----------
        key_path    string   of form 'a/b/c/../q' where all but last
                             component must be a dict

        Returns
        -------
        Value associated with key_path if config contains such a path
        '''
        path_items = key_path.split('/')
        d = self._cfg
        for i in path_items[:-1]:
            if not i in d:
                raise ValueError(f'Item {i} not found')
            d = d[i]
            if not isinstance(d, dict):
                raise ValueError(f'intermediate {d} is not a dict')
        return d[path_items[-1]]

    def validate(self, schema=None):
        '''
        Parameters
        ----------
        schema    Identifies schema to validate against.
                  If string, open file with this file path
                  If dict, use "as is"
                  If None, attempt to deduce schema version (and hence
                  file path) from schema_version keyword
        Returns:  None
                  Raise exception if validation fails for any reason:
                  desc.skycatalogs.exception.NoSchema if schema specification
                  can't be found
                  OSError if schema file can't be read
                  yaml.YAMLerror if schema can't be loaded
                  jsonschema.exceptions.SchemaError if schema is invalid
                  jsonschema.exceptions.ValidationError otherwise
        '''
        fpath = None
        if schema is None:
            if 'schema_version' not in self._cfg:
                raise NoSchemaVersionError
            fpath = _find_schema_path(self._cfg["schema_version"])
            if not fpath:
                raise NoSchemaVersionError('Schema file not found')
        elif isinstance(schema, string):
            fpath = schema
        if fpath:
            try:
                f = open(fpath)
                sch = yaml.safe_load(f)
            except OSError as e:
                raise NoSchemaVersionError('Schema file not found or unreadable')
            except yaml.YAMLError as ye:
                raise ye
        if isinstance(schema, dict):
            sch = schema

        jsonschema.validate(self._cfg, schema.dict)

    def add_key(self, k, v):
        '''
        Parameters
        ----------
        k        Top-level key belonging to the schema
        v        value for the key
        '''

        if k in self._cfg:
            raise ConfigDuplicateKeyError(k)
        self._cfg[k] = v

    def write_config(self, dirpath, filename=None, overwrite=False):
        '''
        Export self to yaml document and write to specified directory.
        If filename is None the file will be named after catalog_name

        Parameters
        ----------
        dirpath        Directory to which file will be written
        filename       If supplied, use for filename
        overwrite      By default do not overwrite existing file with same path

        Return
        ------
        Full path of output config
        '''
        ###self.validate()   skip for now

        if not filename:
            filename = self._cfg['catalog_name'] + '.yaml'

        return write_yaml(self._cfg, os.path.join(dirpath, filename),
                          overwrite=overwrite, logname=self._logname)


def write_yaml(input_dict, outpath, overwrite=False, logname=None):
        if not overwrite:
            try:
                with open(outpath, mode='x') as f:
                    yaml.dump(input_dict, f)
            except FileExistsError:
                if logname:
                    logger = logging.getLogger(logname)
                    logger.warning('Config.write_yaml: Will not overwrite pre-existing config file ' + outpath)
                else:
                    print('Config.write_yaml: Will not overwrite pre-existing config file ' + outpath)
                return

        else:
            with open(outpath, mode='w') as f:
                yaml.dump(input_dict, f)

        return outpath


def _find_schema_path(schema_spec):
    '''
    Given a schema version specification, return the file path
    where the file describing it belongs
    '''
    fname = f'skycatalogs_schema_{self._cfg["schema_spec"]}'
    here = os.path.dirname(__file__)
    return os.path.join(here, '../../../../cfg', fname)

def create_config(catalog_name, logname=None, schema_version=None):
    if not schema_version:
        schema_version = _CURRENT_SCHEMA_VERSION
    return Config({'catalog_name' : catalog_name,
                   'schema_version' : schema_version}, logname)

def assemble_cosmology(cosmology):
    d = {k : cosmology.__getattribute__(k) for k in ('Om0', 'Ob0', 'sigma8',
                                                     'n_s')}
    d['H0'] = float(cosmology.H0.value)
    return d

def assemble_MW_extinction():
    av = {'mode' : 'data'}
    rv = {'mode' : 'constant', 'value' : 3.1}
    return {'r_v' : rv, 'a_v' : av}

def assemble_object_types(pkg_root):
    '''
    Include all supported object types even though a particular catalog
    might not use them all
    '''
    t_path = os.path.join(pkg_root, 'cfg', 'object_types.yaml')
    with open(t_path) as f:
        d = yaml.safe_load(f)
        return d['object_types']

def assemble_SED_models(bins):
    tophat_d = { 'units' : 'angstrom', 'bin_parameters' : ['start', 'width']}
    tophat_d['bins'] = bins
    file_nm_d = {'units' : 'nm'}
    return {'tophat' : tophat_d, 'file_nm' : file_nm_d}

def assemble_provenance(pkg_root, inputs={}):
    repo = git.Repo(pkg_root)
    has_uncommited = repo.is_dirty()
    has_untracked = (len(repo.untracked_files) > 0)

    git_d = {}
    git_d['git_hash'] = repo.commit().hexsha
    git_d['git_branch'] = repo.active_branch.name
    status = []
    if has_uncommited:
        status.append('UNCOMMITTED_FILES')
    if has_untracked:
        status.append('UNTRACKED_FILES')
    if len(status) == 0:
        status.append('CLEAN')
    git_d['git_status'] = status

    if inputs:
        return {'skyCatalogs_repo' : git_d, 'inputs' : inputs}
    else:
        return{'skyCatalogs_repo' : git_d}
